# Copyright 2014 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import filecmp
import os
import re
import stat
import subprocess
import sys
import zipfile
from contextlib import contextmanager
from zipfile import ZipFile

import pytest

from pex.common import CopyMode, open_zip, safe_open, temporary_dir, touch
from pex.compatibility import WINDOWS, commonpath
from pex.executor import Executor
from pex.layout import Layout
from pex.pex import PEX
from pex.pex_builder import Check, InvalidZipAppError, PEXBuilder
from pex.pex_warnings import PEXWarning
from pex.typing import TYPE_CHECKING
from pex.variables import ENV
from testing import IS_PYPY, NonDeterministicWalk, WheelBuilder, install_wheel, make_bdist, make_env
from testing import write_simple_pex as write_pex

try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore[no-redef,import]

if TYPE_CHECKING:
    from typing import Any, Iterator, List, Set, Text

exe_main = """
import sys
from p1.my_module import do_something
do_something()

with open(sys.argv[1], 'w') as fp:
  fp.write('success')
"""


def test_pex_builder():
    # type: () -> None
    with temporary_dir() as td, make_bdist("p1") as p1:
        pb = write_pex(td, exe_main, dists=[p1])

        success_txt = os.path.join(td, "success.txt")
        PEX(td, interpreter=pb.interpreter).run(args=[success_txt])
        assert os.path.exists(success_txt)
        with open(success_txt) as fp:
            assert fp.read() == "success"


def test_pex_builder_shebang():
    # type: () -> None
    def builder(shebang):
        # type: (str) -> PEXBuilder
        pb = PEXBuilder()
        pb.set_shebang(shebang)
        return pb

    for pb in builder("foobar"), builder("#!foobar"):
        for b in pb, pb.clone():
            with temporary_dir() as td:
                target = os.path.join(td, "foo.pex")
                b.build(target)
                expected_preamble = b"#!foobar\n"
                with open(target, "rb") as fp:
                    assert fp.read(len(expected_preamble)) == expected_preamble


def test_pex_builder_preamble():
    # type: () -> None
    with temporary_dir() as td:
        target = os.path.join(td, "foo.pex")
        should_create = os.path.join(td, "foo.1")

        tempfile_preamble = "\n".join(
            ["import sys", "open('{0}', 'w').close()".format(should_create), "sys.exit(3)"]
        )

        pb = PEXBuilder(preamble=tempfile_preamble)
        pb.build(target)

        assert not os.path.exists(should_create)

        pex = PEX(target, interpreter=pb.interpreter)
        process = pex.run(blocking=False)
        process.wait()

        assert process.returncode == 3
        assert os.path.exists(should_create)


def test_pex_builder_compilation():
    # type: () -> None
    with temporary_dir() as td1, temporary_dir() as td2, temporary_dir() as td3:
        src = os.path.join(td1, "src.py")
        with open(src, "w") as fp:
            fp.write(exe_main)

        exe = os.path.join(td1, "exe.py")
        with open(exe, "w") as fp:
            fp.write(exe_main)

        def build_and_check(path, precompile):
            # type: (str, bool) -> None
            pb = PEXBuilder(path=path)
            pb.add_source(src, "lib/src.py")
            pb.set_executable(exe, "exe.py")
            pb.freeze(bytecode_compile=precompile)
            for pyc_file in ("exe.pyc", "lib/src.pyc", "__main__.pyc"):
                pyc_exists = os.path.exists(os.path.join(path, pyc_file))
                if precompile:
                    assert pyc_exists
                else:
                    assert not pyc_exists
            bootstrap_dir = os.path.join(path, pb.info.bootstrap)
            bootstrap_pycs = []  # type: List[str]
            for _, _, files in os.walk(bootstrap_dir):
                bootstrap_pycs.extend(f for f in files if f.endswith(".pyc"))
            if precompile:
                assert len(bootstrap_pycs) > 0
            else:
                assert 0 == len(bootstrap_pycs)

        build_and_check(td2, False)
        build_and_check(td3, True)


@pytest.mark.skipif(WINDOWS, reason="No hardlinks on windows")
def test_pex_builder_copy_or_link():
    # type: () -> None
    with temporary_dir() as td:
        src = os.path.join(td, "exe.py")
        with safe_open(src, "w") as fp:
            fp.write(exe_main)

        def build_and_check(copy_mode):
            # type: (CopyMode.Value) -> None
            pb = PEXBuilder(copy_mode=copy_mode)
            path = pb.path()
            pb.add_source(src, "exe.py")

            path_clone = os.path.join(path, "__clone")
            pb.clone(into=path_clone)

            for root in path, path_clone:
                s1 = os.stat(src)
                s2 = os.stat(os.path.join(root, "exe.py"))
                is_link = (s1[stat.ST_INO], s1[stat.ST_DEV]) == (s2[stat.ST_INO], s2[stat.ST_DEV])
                if copy_mode == CopyMode.COPY:
                    assert not is_link
                else:
                    # Since os.stat follows symlinks; so in CopyMode.SYMLINK, this just proves
                    # the symlink points to the original file. Going further and checking path
                    # and path_clone for the presence of a symlink (an os.islink test) is
                    # trickier since a Linux hardlink of a symlink produces a symlink whereas a
                    # macOS hardlink of a symlink produces a hardlink.
                    assert is_link

        build_and_check(CopyMode.LINK)
        build_and_check(CopyMode.COPY)
        build_and_check(CopyMode.SYMLINK)


@pytest.fixture
def tmp_chroot(tmpdir):
    # type: (Any) -> Iterator[str]
    tmp_chroot = str(tmpdir)
    cwd = os.getcwd()
    try:
        os.chdir(tmp_chroot)
        yield tmp_chroot
    finally:
        os.chdir(cwd)


@pytest.mark.parametrize(
    "copy_mode", [pytest.param(copy_mode, id=copy_mode.value) for copy_mode in CopyMode.values()]
)
def test_pex_builder_add_source_relpath_issues_1192(
    tmp_chroot,  # type: str
    copy_mode,  # type: CopyMode.Value
):
    # type: (...) -> None
    pb = PEXBuilder(copy_mode=copy_mode)
    with safe_open("src/main.py", "w") as fp:
        fp.write("import sys; sys.exit(42)")
    pb.add_source("src/main.py", "main.py")
    pb.set_entry_point("main")
    pb.build("test.pex")

    process = Executor.open_process(cmd=[os.path.abspath("test.pex")])
    process.wait()
    assert 42 == process.returncode


def test_pex_builder_deterministic_timestamp():
    # type: () -> None
    pb = PEXBuilder()
    with temporary_dir() as td:
        target = os.path.join(td, "foo.pex")
        pb.build(target, deterministic_timestamp=True)
        with zipfile.ZipFile(target) as zf:
            assert all(zinfo.date_time == (1980, 1, 1, 0, 0, 0) for zinfo in zf.infolist())


def test_pex_builder_script_from_pex_path(tmpdir):
    # type: (Any) -> None

    pex_with_script = os.path.join(str(tmpdir), "script.pex")
    with make_bdist(
        name="my_project",
        entry_points={"console_scripts": ["my_app = my_project.my_module:do_something"]},
    ) as dist:
        pb = PEXBuilder()
        pb.add_dist_location(dist.location)
        pb.build(pex_with_script)

    pex_file = os.path.join(str(tmpdir), "app.pex")
    pb = PEXBuilder()
    pb.info.pex_path = [pex_with_script]
    pb.set_script("my_app")
    pb.build(pex_file)

    assert "hello world!\n" == subprocess.check_output(args=[pex_file]).decode("utf-8")


def test_pex_builder_setuptools_script(tmpdir):
    # type: (Any) -> None

    pex_file = os.path.join(str(tmpdir), "app.pex")
    with make_bdist(
        name="my_project",
    ) as dist:
        pb = PEXBuilder()
        pb.add_dist_location(dist.location)
        pb.set_script("shell_script")
        pb.build(pex_file)

    assert "hello world from shell script\n" == subprocess.check_output(args=[pex_file]).decode(
        "utf-8"
    )


def test_pex_builder_packed(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    pex_app = os.path.join(str(tmpdir), "app.pex")
    source_file = os.path.join(str(tmpdir), "src")
    touch(source_file)

    with ENV.patch(PEX_ROOT=pex_root), make_bdist(name="my_project") as dist:
        pb = PEXBuilder(copy_mode=CopyMode.SYMLINK)
        pb.add_source(source_file, "a.file")
        pb.add_dist_location(dist.location)
        pb.set_script("shell_script")
        pb.build(pex_app, layout=Layout.PACKED)

    assert "hello world from shell script\n" == subprocess.check_output(
        args=[os.path.join(pex_app, "__main__.py")]
    ).decode("utf-8")

    spread_dist_bootstrap = os.path.join(pex_app, pb.info.bootstrap)
    assert zipfile.is_zipfile(spread_dist_bootstrap)

    cached_bootstrap_zip = os.path.join(
        pex_root, "bootstrap_zips", pb.info.bootstrap_hash, pb.info.bootstrap
    )
    assert zipfile.is_zipfile(cached_bootstrap_zip)

    assert filecmp.cmp(spread_dist_bootstrap, cached_bootstrap_zip, shallow=False)

    assert os.path.isfile(os.path.join(pex_app, "a.file"))
    for root, dirs, files in os.walk(pex_app, followlinks=False):
        for f in files:
            path = os.path.join(root, f)
            assert not os.path.islink(path) or pex_app == commonpath(
                [pex_app, os.path.realpath(path)]
            ), (
                "All packed layout files should be real files inside the packed layout root that "
                "are divorced from either the PEXBuilder chroot or PEX_ROOT caches."
            )

    assert 1 == len(pb.info.distributions)
    location, sha = next(iter(pb.info.distributions.items()))

    spread_dist_zip = os.path.join(pex_app, pb.info.internal_cache, location)
    assert zipfile.is_zipfile(spread_dist_zip)

    cached_dist_zip = os.path.join(pex_root, "packed_wheels", sha, location)
    assert zipfile.is_zipfile(cached_dist_zip)

    assert filecmp.cmp(spread_dist_zip, cached_dist_zip, shallow=False)


@pytest.mark.parametrize(
    "copy_mode", [pytest.param(copy_mode, id=copy_mode.value) for copy_mode in CopyMode.values()]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
def test_pex_builder_exclude_bootstrap_testing(
    tmpdir,  # type: Any
    copy_mode,  # type: CopyMode.Value
    layout,  # type: Layout.Value
):
    # type: (...) -> None

    pex_path = os.path.join(str(tmpdir), "empty.pex")
    pb = PEXBuilder(copy_mode=copy_mode)
    pb.build(pex_path, layout=layout)

    bootstrap_location = os.path.join(pex_path, pb.info.bootstrap)
    bootstrap_files = set()  # type: Set[Text]
    if Layout.ZIPAPP == layout:
        with open_zip(pex_path) as zf:
            bootstrap_files.update(
                os.path.relpath(f, pb.info.bootstrap)
                for f in zf.namelist()
                if f.startswith(pb.info.bootstrap)
            )
    elif Layout.PACKED == layout:
        with open_zip(bootstrap_location) as zf:
            bootstrap_files.update(zf.namelist())
    else:
        bootstrap_files.update(
            os.path.relpath(os.path.join(root, f), bootstrap_location)
            for root, _, files in os.walk(bootstrap_location)
            for f in files
        )

    assert {"pex/pex_bootstrapper.py", "pex/pex_info.py", "pex/pex.py"}.issubset(
        bootstrap_files
    ), "Expected the `.bootstrap` to contain at least some of the key Pex runtime modules."
    assert not [
        f for f in bootstrap_files if f.endswith(("testing.py", "testing.pyc"))
    ], "Expected testing support files to be stripped from the Pex `.bootstrap`."


def test_pex_builder_deterministic_pex_info(
    tmpdir,  # type: Any
):
    # type: (...) -> None

    pex_path = os.path.join(str(tmpdir), "empty.pex")

    with mock.patch("os.walk", new=NonDeterministicWalk()):
        pb_one = PEXBuilder()
        pb_one.build(pex_path)

        pb_two = PEXBuilder()
        pb_two.build(pex_path)

        assert pb_one.info.as_json_dict() == pb_two.info.as_json_dict()


@pytest.mark.parametrize(
    "strip_pex_env", [pytest.param(True, id="StripPexEnv"), pytest.param(False, id="NoStripPexEnv")]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize("venv", [pytest.param(True, id="VENV"), pytest.param(False, id="UNZIP")])
def test_pex_env_var_issues_1485(
    tmpdir,  # type: Any
    strip_pex_env,  # type: bool
    venv,  # type: bool
    layout,  # type: Layout.Value
):
    # type: (...) -> None

    pex_path = os.path.join(str(tmpdir), "empty.pex")
    pb = PEXBuilder()
    pb.info.strip_pex_env = strip_pex_env
    pb.info.venv = venv
    pb.build(pex_path, layout=layout)

    launch_args = [pex_path] if layout == Layout.ZIPAPP else [sys.executable, pex_path]
    pex_root = os.path.join(str(tmpdir), "pex_root")

    def assert_pex_env_var(
        script,  # type: str
        expected_pex_env_var,  # type: str
        expect_empty_pex_root,  # type: bool
        expected_layout=layout,  # type: Layout.Value
    ):
        # type: (...) -> None
        if expect_empty_pex_root:
            assert not os.path.exists(pex_root)
        else:
            assert len(os.listdir(pex_root)) > 0
        output = subprocess.check_output(
            launch_args + ["-c", script], env=make_env(PEX_ROOT=pex_root)
        )
        actual = output.decode("utf-8").strip()
        assert os.path.realpath(expected_pex_env_var) == actual
        if Layout.ZIPAPP == expected_layout:
            assert zipfile.is_zipfile(actual)
        else:
            assert os.path.isdir(actual)

    print_pex_env_var_script = "import os; print(os.environ['PEX'])"

    assert_pex_env_var(
        script=print_pex_env_var_script,
        expected_pex_env_var=pex_path,
        expect_empty_pex_root=True,
    )
    assert_pex_env_var(
        script=print_pex_env_var_script,
        expected_pex_env_var=pex_path,
        expect_empty_pex_root=False,
    )

    other_pex_path = os.path.join(str(tmpdir), "other.pex")
    other_pb = PEXBuilder()
    other_pb.info.includes_tools = True
    other_pex_main = os.path.join(str(tmpdir), "main.py")
    with open(other_pex_main, mode="w") as fp:
        fp.write(print_pex_env_var_script)
    other_pb.set_executable(other_pex_main)
    other_pb.build(other_pex_path, layout=Layout.ZIPAPP)

    def assert_pex_env_var_nested(**env):
        # type: (**Any) -> None
        assert_pex_env_var(
            script="import subprocess; subprocess.check_call([{other_pex!r}], env={env!r})".format(
                other_pex=other_pex_path, env=make_env(**env)
            ),
            expected_pex_env_var=other_pex_path,
            expect_empty_pex_root=False,
            expected_layout=Layout.ZIPAPP,
        )

    assert_pex_env_var_nested()
    assert_pex_env_var_nested(PEX_VENV=False)
    assert_pex_env_var_nested(PEX_VENV=True)


@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in (Layout.PACKED, Layout.ZIPAPP)]
)
def test_build_compression(
    tmpdir,  # type: Any
    layout,  # type: Layout.Value
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    pb = PEXBuilder()
    pb.info.pex_root = pex_root
    with ENV.patch(PEX_ROOT=pex_root):
        pb.add_dist_location(install_wheel(WheelBuilder(pex_project_dir).bdist()).location)
    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write("import pex; print(pex.__file__)")
    pb.set_executable(exe)

    def assert_pex(pex):
        # type: (str) -> None
        assert (
            subprocess.check_output(args=[sys.executable, pex]).decode("utf-8").startswith(pex_root)
        )

    compressed_pex = os.path.join(str(tmpdir), "compressed.pex")
    pb.build(compressed_pex, layout=layout)
    assert_pex(compressed_pex)

    uncompressed_pex = os.path.join(str(tmpdir), "uncompressed.pex")
    pb.build(uncompressed_pex, layout=layout, compress=False)
    assert_pex(uncompressed_pex)

    def size(pex):
        # type: (str) -> int
        if os.path.isfile(pex):
            return os.path.getsize(pex)
        return sum(
            os.path.getsize(os.path.join(root, f)) for root, _, files in os.walk(pex) for f in files
        )

    assert size(compressed_pex) < size(uncompressed_pex)


@pytest.mark.skipif(
    sys.version_info == (3, 13, 0, "alpha", 6),
    reason=(
        "The 3.13.0a6 release has a bug in its zipimporter ZIP64 handling, see: "
        "https://github.com/python/cpython/issues/118107"
    ),
)
@pytest.mark.skipif(
    subprocess.call(args=["cat", os.devnull]) != 0,
    reason="The cat binary is required for this test.",
)
def test_check(tmpdir):
    # type: (Any) -> None

    def assert_perform_check_zip64_handling(
        zipapp,  # type: str
        test_run=True,  # type: bool
    ):
        # type: (...) -> None
        assert Check.NONE.perform_check(Layout.ZIPAPP, zipapp) is None
        assert Check.ERROR.perform_check(Layout.PACKED, zipapp) is None
        assert Check.ERROR.perform_check(Layout.LOOSE, zipapp) is None

        # Python 3.13 newly supports ZIP64 in `zipimport.zipimporter`.
        if sys.version_info >= (3, 13):
            assert Check.WARN.perform_check(Layout.ZIPAPP, zipapp) is True
            assert Check.ERROR.perform_check(Layout.ZIPAPP, zipapp) is True
            if test_run:
                subprocess.check_call(args=[sys.executable, zipapp])
        else:
            expected_error_message_re = r"The PEX zip at {path} is not a valid zipapp: ".format(
                path=re.escape(zipapp)
            )
            with pytest.warns(PEXWarning, match=expected_error_message_re):
                assert Check.WARN.perform_check(Layout.ZIPAPP, zipapp) is False
            with pytest.raises(InvalidZipAppError, match=expected_error_message_re):
                Check.ERROR.perform_check(Layout.ZIPAPP, zipapp)

            if test_run:
                assert subprocess.call(args=[sys.executable, zipapp]) != 0

    @contextmanager
    def write_zipapp(path):
        # type: (str) -> Iterator[ZipFile]
        with open_zip(path, "w") as zip_file:
            yield zip_file
            zip_file.writestr("__main__.py", "print('BOOTED')")

    file_too_big = os.path.join(str(tmpdir), "file_too_big.py")
    with open(file_too_big, "w") as fp:
        fp.write("\n")
    for _ in range(32):
        accum = file_too_big + ".accum"
        os.rename(file_too_big, accum)
        with open(file_too_big, "wb") as dest:
            subprocess.check_call(args=["cat", accum, accum], stdout=dest.fileno())
    assert os.path.getsize(file_too_big) > 0xFFFFFFFF, (
        "ZIP64 must be used if file sizes are bigger than 0xFFFFFFFF per "
        "https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT 4.3.9.2"
    )

    zipapp_too_big = os.path.join(str(tmpdir), "too-big.pyz")
    with write_zipapp(zipapp_too_big) as zf:
        zf.write(file_too_big, os.path.basename(file_too_big))
    assert_perform_check_zip64_handling(
        zipapp_too_big,
        # N.B.: PyPy < 3.8 hangs when trying to run a zip64 app with a too-large file entry.
        test_run=not IS_PYPY or sys.version_info[:2] >= (3, 8),
    )

    zipapp_too_many_entries = os.path.join(str(tmpdir), "too-many-entries.pyz")
    with write_zipapp(zipapp_too_many_entries) as zf:
        # N.B.: ZIP64 must be used if the number of central directory records is greater than
        # 0xFFFF per https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT 4.4.21
        for x in range(0xFFFF):
            zf.writestr(
                "module{x}.py".format(x=x),
                "# This is python module number {x}.".format(x=x).encode("utf-8"),
            )
    assert_perform_check_zip64_handling(zipapp_too_many_entries)

    zipapp_ok = os.path.join(str(tmpdir), "ok.pyz")
    with write_zipapp(zipapp_ok):
        pass
    for layout in Layout.values():
        for check in Check.values():
            assert check.perform_check(layout, zipapp_ok) is (
                True if layout is Layout.ZIPAPP and check is not Check.NONE else None
            )
    assert b"BOOTED\n" == subprocess.check_output(args=[sys.executable, zipapp_ok])



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import pex_build
from hatchling.plugin import hookimpl
from pex_build.hatchling.build_hook import AdjustBuild
from pex_build.hatchling.metadata_hook import AdjustMetadata

if pex_build.TYPE_CHECKING:
    from typing import Type


@hookimpl
def hatch_register_metadata_hook():
    # type: () -> Type[AdjustMetadata]
    return AdjustMetadata


@hookimpl
def hatch_register_build_hook():
    # type: () -> Type[AdjustBuild]
    return AdjustBuild



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import os
import platform
from collections import defaultdict

from pex.enum import Enum
from pex.platforms import Platform
from pex.targets import Targets
from pex.third_party.packaging import tags  # noqa
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import DefaultDict, Iterable, Optional, Set, Tuple, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class ScieStyle(Enum["ScieStyle.Value"]):
    class Value(Enum.Value):
        pass

    LAZY = Value("lazy")
    EAGER = Value("eager")


class _CurrentPlatform(object):
    def __get__(self, obj, objtype=None):
        # type: (...) -> SciePlatform.Value
        if not hasattr(self, "_current"):
            system = platform.system().lower()
            machine = platform.machine().lower()
            if "linux" == system:
                if machine in ("aarch64", "arm64"):
                    self._current = SciePlatform.LINUX_AARCH64
                elif machine in ("amd64", "x86_64"):
                    self._current = SciePlatform.LINUX_X86_64
            elif "darwin" == system:
                if machine in ("aarch64", "arm64"):
                    self._current = SciePlatform.MACOS_AARCH64
                elif machine in ("amd64", "x86_64"):
                    self._current = SciePlatform.MACOS_X86_64
            elif "windows" == system:
                if machine in ("aarch64", "arm64"):
                    self._current = SciePlatform.WINDOWS_AARCH64
                elif machine in ("amd64", "x86_64"):
                    self._current = SciePlatform.WINDOWS_X86_64
            if not hasattr(self, "_current"):
                raise ValueError(
                    "The current operating system / machine pair is not supported!: "
                    "{system} / {machine}".format(system=system, machine=machine)
                )
        return self._current


class SciePlatform(Enum["SciePlatform.Value"]):
    class Value(Enum.Value):
        @property
        def extension(self):
            # type: () -> str
            return (
                ".exe"
                if self in (SciePlatform.WINDOWS_AARCH64, SciePlatform.WINDOWS_X86_64)
                else ""
            )

        def binary_name(self, binary_name):
            # type: (str) -> str
            return "{binary_name}{extension}".format(
                binary_name=binary_name, extension=self.extension
            )

        def qualified_binary_name(self, binary_name):
            # type: (str) -> str
            return "{binary_name}-{platform}{extension}".format(
                binary_name=binary_name, platform=self, extension=self.extension
            )

        def qualified_file_name(self, file_name):
            # type: (str) -> str
            stem, ext = os.path.splitext(file_name)
            return "{stem}-{platform}{ext}".format(stem=stem, platform=self, ext=ext)

    LINUX_AARCH64 = Value("linux-aarch64")
    LINUX_X86_64 = Value("linux-x86_64")
    MACOS_AARCH64 = Value("macos-aarch64")
    MACOS_X86_64 = Value("macos-x86_64")
    WINDOWS_AARCH64 = Value("windows-x86_64")
    WINDOWS_X86_64 = Value("windows-aarch64")
    CURRENT = _CurrentPlatform()

    @classmethod
    def parse(cls, value):
        # type: (str) -> SciePlatform.Value
        return cls.CURRENT if "current" == value else cls.for_value(value)


@attr.s(frozen=True)
class ScieTarget(object):
    platform = attr.ib()  # type: SciePlatform.Value
    python_version = attr.ib()  # type: Union[Tuple[int, int], Tuple[int, int, int]]
    pbs_release = attr.ib(default=None)  # type: Optional[str]

    @property
    def version_str(self):
        # type: () -> str
        return ".".join(map(str, self.python_version))


@attr.s(frozen=True)
class ScieInfo(object):
    style = attr.ib()  # type: ScieStyle.Value
    target = attr.ib()  # type: ScieTarget
    file = attr.ib()  # type: str

    @property
    def platform(self):
        # type: () -> SciePlatform.Value
        return self.target.platform

    @property
    def python_version(self):
        # type: () -> Union[Tuple[int, int], Tuple[int, int, int]]
        return self.target.python_version


@attr.s(frozen=True)
class ScieOptions(object):
    style = attr.ib(default=ScieStyle.LAZY)  # type: ScieStyle.Value
    platforms = attr.ib(default=())  # type: Tuple[SciePlatform.Value, ...]
    pbs_release = attr.ib(default=None)  # type: Optional[str]
    python_version = attr.ib(
        default=None
    )  # type: Optional[Union[Tuple[int, int], Tuple[int, int, int]]]
    science_binary_url = attr.ib(default=None)  # type: Optional[str]

    def create_configuration(self, targets):
        # type: (Targets) -> ScieConfiguration
        return ScieConfiguration.from_targets(self, targets)


@attr.s(frozen=True)
class ScieConfiguration(object):
    @classmethod
    def from_tags(
        cls,
        options,  # type: ScieOptions
        tags,  # type: Iterable[tags.Tag]
    ):
        # type: (...) -> ScieConfiguration
        return cls._from_platforms(
            options=options, platforms=tuple(Platform.from_tag(tag) for tag in tags)
        )

    @classmethod
    def from_targets(
        cls,
        options,  # type: ScieOptions
        targets,  # type: Targets
    ):
        # type: (...) -> ScieConfiguration
        return cls._from_platforms(
            options=options,
            platforms=tuple(target.platform for target in targets.unique_targets()),
        )

    @classmethod
    def _from_platforms(
        cls,
        options,  # type: ScieOptions
        platforms,  # type: Iterable[Platform]
    ):
        # type: (...) -> ScieConfiguration

        python_version = options.python_version
        python_versions_by_platform = defaultdict(
            set
        )  # type: DefaultDict[SciePlatform.Value, Set[Union[Tuple[int, int], Tuple[int, int, int]]]]
        for plat in platforms:
            if python_version:
                plat_python_version = python_version
            elif len(plat.version_info) < 2:
                continue
            else:
                # Here were guessing an available PBS CPython version. Since a triple is unlikely to
                # hit, we just use major / minor. If the user wants control they can specify
                # options.python_version via `--scie-python-version`.
                plat_python_version = cast(
                    "Union[Tuple[int, int], Tuple[int, int, int]]", plat.version_info
                )[:2]

            # We use Python Build Standalone to create scies, and we know it does not support
            # CPython<3.8.
            if plat_python_version < (3, 8):
                continue

            # We use Python Build Standalone to create scies, and we know it only provides CPython
            # interpreters.
            if plat.impl not in ("py", "cp"):
                continue

            platform_str = plat.platform
            is_aarch64 = "arm64" in platform_str or "aarch64" in platform_str
            is_x86_64 = "amd64" in platform_str or "x86_64" in platform_str
            if not is_aarch64 ^ is_x86_64:
                continue

            if "linux" in platform_str:
                scie_platform = (
                    SciePlatform.LINUX_AARCH64 if is_aarch64 else SciePlatform.LINUX_X86_64
                )
            elif "mac" in platform_str:
                scie_platform = (
                    SciePlatform.MACOS_AARCH64 if is_aarch64 else SciePlatform.MACOS_X86_64
                )
            elif "win" in platform_str:
                scie_platform = (
                    SciePlatform.WINDOWS_AARCH64 if is_aarch64 else SciePlatform.WINDOWS_X86_64
                )
            else:
                continue

            python_versions_by_platform[scie_platform].add(plat_python_version)

        for explicit_platform in options.platforms:
            if explicit_platform not in python_versions_by_platform:
                if options.python_version:
                    python_versions_by_platform[explicit_platform] = {options.python_version}
                else:
                    python_versions_by_platform[explicit_platform] = set(
                        itertools.chain.from_iterable(python_versions_by_platform.values())
                    )
        if options.platforms:
            for configured_platform in tuple(python_versions_by_platform):
                if configured_platform not in options.platforms:
                    python_versions_by_platform.pop(configured_platform, None)

        scie_targets = tuple(
            ScieTarget(
                platform=scie_platform,
                pbs_release=options.pbs_release,
                python_version=max(python_versions),
            )
            for scie_platform, python_versions in sorted(python_versions_by_platform.items())
        )
        return cls(options=options, targets=tuple(scie_targets))

    options = attr.ib()  # type: ScieOptions
    targets = attr.ib()  # type: Tuple[ScieTarget, ...]

    def __len__(self):
        # type: () -> int
        return len(self.targets)



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
import shutil
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.typing import TYPE_CHECKING
from testing import IS_PYPY, run_pex_command

if TYPE_CHECKING:
    from typing import Any, Iterable, List

parametrize_execution_mode_args = pytest.mark.parametrize(
    "execution_mode_args",
    [
        pytest.param([], id="ZIPAPP"),
        pytest.param(["--venv"], id="VENV"),
    ],
)


parametrize_boot_mode_args = pytest.mark.parametrize(
    "boot_mode_args",
    [
        pytest.param([], id="PYTHON"),
        pytest.param(["--sh-boot"], id="SH_BOOT"),
    ],
)


@pytest.fixture
def exe(tmpdir):
    # type: (Any) -> str
    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import sys
                import warnings


                warnings.warn("If you don't eat your meat, you can't have any pudding!")
                json.dump(sys.argv[1:], sys.stdout)
                """
            )
        )
    return exe


def assert_exe_output(
    pex,  # type: str
    warning_expected,  # type: bool
    prefix_args=(),  # type: Iterable[str]
):
    process = subprocess.Popen(
        args=list(prefix_args) + [pex, "--foo", "bar"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    error = stderr.decode("utf-8")
    assert 0 == process.returncode, error
    assert ["--foo", "bar"] == json.loads(stdout)
    assert warning_expected == (
        "If you don't eat your meat, you can't have any pudding!" in error
    ), error


@pytest.mark.skipif(
    IS_PYPY and sys.version_info[:2] < (3, 10),
    reason=(
        "Pex cannot retrieve the original argv when running under PyPy<3.10 which prevents "
        "passthrough."
    ),
)
@parametrize_execution_mode_args
@parametrize_boot_mode_args
def test_python_args_passthrough(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
    boot_mode_args,  # type: List[str]
    exe,  # type: str
):
    # type: (...) -> None

    default_shebang_pex = os.path.join(str(tmpdir), "default_shebang.pex")
    custom_shebang_pex = os.path.join(str(tmpdir), "custom_shebang.pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")

    args = (
        execution_mode_args
        + boot_mode_args
        + [
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--exe",
            exe,
        ]
    )
    run_pex_command(args=args + ["-o", default_shebang_pex]).assert_success()
    run_pex_command(
        args=args
        + [
            "-o",
            custom_shebang_pex,
            "--python-shebang",
            "{python} -Wignore".format(python=sys.executable),
        ]
    ).assert_success()

    # N.B.: We execute tests in doubles after a cache nuke to exercise both cold and warm runs
    # which take different re-exec paths through the code that all need to preserve Python args.

    # The built-in python shebang args, if any, should be respected.
    shutil.rmtree(pex_root)
    assert_exe_output(default_shebang_pex, warning_expected=True)
    assert_exe_output(default_shebang_pex, warning_expected=True)
    assert_exe_output(custom_shebang_pex, warning_expected=False)
    assert_exe_output(custom_shebang_pex, warning_expected=False)

    # But they also should be able to be over-ridden.
    shutil.rmtree(pex_root)
    assert_exe_output(
        default_shebang_pex, prefix_args=[sys.executable, "-Wignore"], warning_expected=False
    )
    assert_exe_output(
        default_shebang_pex, prefix_args=[sys.executable, "-Wignore"], warning_expected=False
    )
    assert_exe_output(custom_shebang_pex, prefix_args=[sys.executable], warning_expected=True)
    assert_exe_output(custom_shebang_pex, prefix_args=[sys.executable], warning_expected=True)


@parametrize_execution_mode_args
@parametrize_boot_mode_args
def test_inject_python_args(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
    boot_mode_args,  # type: List[str]
    exe,  # type: str
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")

    run_pex_command(
        args=execution_mode_args
        + boot_mode_args
        + [
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--exe",
            exe,
            "--inject-python-args=-W ignore",
            "-o",
            pex,
        ]
    ).assert_success()

    assert_exe_output(pex, warning_expected=False)
    assert_exe_output(pex, warning_expected=False)

    # N.B.: The original argv cannot be detected by Pex running under PyPy<3.10; so we expect
    # warnings to be turned off (the default sealed in by `--inject-python-args`). For all other
    # Pythons we support, these explicit command line Python args should be detected and trump the
    # injected args by dint of occurring later in the command line. In other words, the command line
    # should be as follows and Python is known to pick the last occurrence of the -W option:
    #
    #   python -W ignore -W always ...
    #
    warning_expected = not IS_PYPY or sys.version_info[:2] >= (3, 10)
    assert_exe_output(
        pex, prefix_args=[sys.executable, "-W", "always"], warning_expected=warning_expected
    )
    assert_exe_output(
        pex, prefix_args=[sys.executable, "-W", "always"], warning_expected=warning_expected
    )


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 10 if IS_PYPY else 9),
    reason=(
        "The effect of `-u` on the `sys.stdout.buffer` type used in this test is only "
        "differentiable from a lack of `-u` for these Pythons."
    ),
)
@parametrize_execution_mode_args
@parametrize_boot_mode_args
def test_issue_2422(
    tmpdir,  # type: Any
    execution_mode_args,  # type: List[str]
    boot_mode_args,  # type: List[str]
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    pex_root = os.path.join(str(tmpdir), "pex_root")

    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                import sys


                if __name__ == "__main__":
                    print(type(sys.stdout.buffer).__name__)
                """
            )
        )

    args = ["--pex-root", pex_root, "--runtime-pex-root", pex_root, "--exe", exe, "-o", pex]
    args.extend(execution_mode_args)
    args.extend(boot_mode_args)

    run_pex_command(args=args).assert_success()
    shutil.rmtree(pex_root)
    assert b"BufferedWriter\n" == subprocess.check_output(args=[sys.executable, exe])
    assert b"BufferedWriter\n" == subprocess.check_output(
        args=[pex]
    ), "Expected cold run to use buffered io."
    assert b"BufferedWriter\n" == subprocess.check_output(
        args=[pex]
    ), "Expected warm run to use buffered io."

    assert b"FileIO\n" == subprocess.check_output(
        args=[sys.executable, "-u", pex]
    ), "Expected explicit Python arguments to be honored."

    run_pex_command(args=args + ["--inject-python-args=-u"]).assert_success()
    shutil.rmtree(pex_root)
    assert b"FileIO\n" == subprocess.check_output(args=[sys.executable, "-u", exe])
    assert b"FileIO\n" == subprocess.check_output(
        args=[pex]
    ), "Expected cold run to use un-buffered io."
    assert b"FileIO\n" == subprocess.check_output(
        args=[pex]
    ), "Expected warm run to use un-buffered io."

    process = subprocess.Popen(
        args=[sys.executable, "-v", pex],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    assert 0 == process.returncode
    assert b"FileIO\n" == stdout, "Expected injected Python arguments to be honored."
    assert b"import " in stderr, "Expected explicit Python arguments to be honored as well."



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
from textwrap import dedent

from pex.compatibility import commonpath
from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any


def test_long_wheel_names(tmpdir):
    # type: (Any) -> None

    # N.B.: The pycryptodome 3.16.0 release has a wheel with a 116 character name in
    # pycryptodome-3.16.0-cp35-abi3-manylinux_2_5_x86_64.manylinux1_x86_64.manylinux_2_12_x86_64.manylinux2010_x86_64.whl
    # that triggers special logic to deal with temporary work dir names being too long for some
    # systems, notably those that house the PEX_ROOT on an eCryptFS volume.
    pex_root = os.path.join(str(tmpdir), "pex_root")
    result = run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "pycryptodome==3.16.0",
            "--",
            "-c",
            dedent(
                """\
                import json
                import sys

                import Crypto


                json.dump({"version": Crypto.__version__, "path": Crypto.__file__}, sys.stdout)
                """
            ),
        ]
    )
    result.assert_success()

    data = json.loads(result.output)
    assert "3.16.0" == data["version"]
    assert pex_root == commonpath((pex_root, data["path"]))



# coding=utf-8
# Copyright 2018 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
import sys
import types


class Bootstrap(object):
    """Supports introspection of the PEX bootstrap code."""

    _INSTANCE = None

    @classmethod
    def locate(cls):
        # type: () -> Bootstrap
        """Locates the active PEX bootstrap.

        :rtype: :class:`Bootstrap`
        """
        if cls._INSTANCE is None:
            bootstrap_path = __file__
            module_import_path = __name__.split(".")

            # For example, our __file__ might be requests.pex/.bootstrap/pex/bootstrap.pyc and our import
            # path pex.bootstrap; so we walk back through all the module components of our import path to
            # find the base sys.path entry where we were found (requests.pex/.bootstrap in this example).
            for _ in module_import_path:
                bootstrap_path = os.path.dirname(bootstrap_path)

            cls._INSTANCE = cls(sys_path_entry=bootstrap_path)
        return cls._INSTANCE

    def __init__(self, sys_path_entry):
        # type: (str) -> None
        self._sys_path_entry = sys_path_entry
        self._realpath = os.path.realpath(self._sys_path_entry)

    @property
    def path(self):
        # type: () -> str
        return self._sys_path_entry

    def demote(self, disable_vendor_importer=True):
        """Demote the bootstrap code to the end of the `sys.path` so it is found last.

        :return: The list of un-imported bootstrap modules.
        :rtype: list of :class:`types.ModuleType`
        """
        import sys  # Grab a hold of `sys` early since we'll be un-importing our module in this process.

        # N.B.: We mutate the sys.path before un-importing modules so that any re-imports triggered
        # by concurrent code will pull from the desired sys.path ordering.
        # See here for how this situation might arise: https://github.com/pex-tool/pex/issues/1272

        sys.path[:] = [path for path in sys.path if os.path.realpath(path) != self._realpath]
        sys.path.append(self._sys_path_entry)

        unimported_modules = []
        for name, module in reversed(sorted(sys.modules.items())):
            if "pex.third_party" == name and not disable_vendor_importer:
                continue
            if self.imported_from_bootstrap(module):
                unimported_modules.append(sys.modules.pop(name))
        return unimported_modules

    def imported_from_bootstrap(self, module):
        """Return ``True`` if the given ``module`` object was imported from bootstrap code.

        :param module: The module to check the provenance of.
        :type module: :class:`types.ModuleType`
        :rtype: bool
        """

        # Python 2.7 does some funky imports in the email stdlib package that cause havoc with
        # un-importing. Since all our own importing just goes through the vanilla importers we can
        # safely ignore all but the standard module type.
        if not isinstance(module, types.ModuleType):
            return False

        # A vendored module.
        path = getattr(module, "__file__", None)
        if path and os.path.realpath(path).startswith(self._realpath):
            return True

        # A vendored package.
        path = getattr(module, "__path__", None)
        if path and any(
            os.path.realpath(path_item).startswith(self._realpath) for path_item in path
        ):
            return True

        return False

    def __repr__(self):
        return "{cls}(sys_path_entry={sys_path_entry!r})".format(
            cls=type(self).__name__, sys_path_entry=self._sys_path_entry
        )


def demote(disable_vendor_importer=True):
    # type: (bool) -> None
    """Demote PEX bootstrap code to the end of `sys.path` and uninstall all PEX vendored code."""

    from . import third_party
    from .tracer import TRACER

    TRACER.log("Bootstrap complete, performing final sys.path modifications...")

    should_log = {level: TRACER.should_log(V=level) for level in range(1, 10)}

    def log(msg, V=1):
        if should_log.get(V, False):
            print("pex: {}".format(msg), file=sys.stderr)

    # Remove the third party resources pex uses and demote pex bootstrap code to the end of
    # sys.path for the duration of the run to allow conflicting versions supplied by user
    # dependencies to win during the course of the execution of user code.
    third_party.uninstall()

    bootstrap = Bootstrap.locate()
    log("Demoting code from %s" % bootstrap, V=2)
    for module in bootstrap.demote(disable_vendor_importer=disable_vendor_importer):
        log("un-imported {}".format(module), V=9)

    import pex

    log("Re-imported pex from {}".format(pex.__path__), V=3)

    log("PYTHONPATH contains:")
    for element in sys.path:
        log("  %c %s" % (" " if os.path.exists(element) else "*", element))
    log("  * - paths that do not exist or will be imported via zipimport")



# Copyright 2020 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from typing import List

from docutils import nodes, statemachine
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.util.nodes import nested_parse_with_titles

from pex.variables import DefaultedProperty, Variables


class Vars(Directive):
    def convert_rst_to_nodes(self, rst_source: str) -> List[nodes.Node]:
        """Turn an RST string into a node that can be used in the document."""
        node = nodes.Element()
        node.document = self.state.document
        nested_parse_with_titles(
            state=self.state, content=statemachine.ViewList(rst_source.split("\n")), node=node
        )
        return node.children

    def run(self) -> List[nodes.Node]:
        def make_nodes(var_name: str) -> List[nodes.Node]:
            var_obj = Variables.__dict__[var_name]
            if isinstance(var_obj, DefaultedProperty):
                desc_str = var_obj._func.__doc__
            else:
                desc_str = var_obj.__doc__
            desc_str = desc_str or "NO DESC"

            sig = addnodes.desc()
            sig["objtype"] = sig["desctype"] = "var"
            sig.append(nodes.target("", "", ids=[var_name]))
            sig.append(addnodes.desc_signature(var_name, var_name))

            var_nodes = [sig]  # type: List[nodes.Node]
            var_nodes.extend(self.convert_rst_to_nodes(desc_str))
            return var_nodes

        return [
            node for var in dir(Variables) if var.startswith("PEX_") for node in make_nodes(var)
        ]



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import hashlib
import os.path
import warnings

import pytest

from pex.dependency_configuration import DependencyConfiguration
from pex.dependency_manager import DependencyManager
from pex.dist_metadata import Distribution, Requirement
from pex.exceptions import reportable_unexpected_error_msg
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.orderedset import OrderedSet
from pex.pep_503 import ProjectName
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.pex_warnings import PEXWarning
from pex.typing import TYPE_CHECKING
from testing.dist_metadata import create_dist_metadata

if TYPE_CHECKING:
    from typing import Any, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class DistFactory(object):
    install_base_dir = attr.ib()  # type: str

    def create(
        self,
        name,  # type: str
        *requires  # type: str
    ):
        # type: (...) -> FingerprintedDistribution
        fingerprint = hashlib.sha256(name.encode("utf-8")).hexdigest()
        location = os.path.join(self.install_base_dir, fingerprint, name)
        os.makedirs(location)
        return FingerprintedDistribution(
            distribution=Distribution(
                location=location,
                metadata=create_dist_metadata(
                    project_name=name, version="0.1.0", requires_dists=requires, location=location
                ),
            ),
            fingerprint=fingerprint,
        )


@pytest.fixture
def dist_factory(tmpdir):
    # type: (Any) -> DistFactory
    return DistFactory(os.path.join(str(tmpdir), "installed_wheels"))


@attr.s(frozen=True)
class DistGraph(object):
    root_reqs = attr.ib()  # type: Tuple[Requirement, ...]
    dists = attr.ib()  # type: Tuple[FingerprintedDistribution, ...]

    def dist(self, name):
        # type: (str) -> FingerprintedDistribution
        project_name = ProjectName(name)
        dists = [
            dist for dist in self.dists if project_name == dist.distribution.metadata.project_name
        ]
        assert len(dists) == 1, "Expected {name} to match one dist, found {found}".format(
            name=name, found=" ".join(map(str, dists)) if dists else "none"
        )
        return dists[0]


@pytest.fixture
def dist_graph(dist_factory):
    # type: (DistFactory) -> DistGraph

    # distA     distB <--------\
    #      \   /     \         |
    #       v v       v        |
    #      distC     distD  (cycle)
    #     /     \   /          |
    #    V       v v           |
    # distE     distF ---------/

    return DistGraph(
        root_reqs=(Requirement.parse("a"), Requirement.parse("b")),
        dists=(
            dist_factory.create("A", "c"),
            dist_factory.create("B", "c", "d"),
            dist_factory.create("C", "e", "f"),
            dist_factory.create("D", "f"),
            dist_factory.create("E"),
            dist_factory.create("F", "b"),
        ),
    )


def test_exclude_root_reqs(dist_graph):
    # type: (DistGraph) -> None

    dependency_manager = DependencyManager(
        requirements=OrderedSet(dist_graph.root_reqs), distributions=OrderedSet(dist_graph.dists)
    )

    pex_info = PexInfo.default()
    pex_builder = PEXBuilder(pex_info=pex_info)

    with warnings.catch_warnings(record=True) as events:
        dependency_manager.configure(
            pex_builder, dependency_configuration=DependencyConfiguration.create(["a", "b"])
        )
    assert 2 == len(events)

    warning = events[0]
    assert PEXWarning == warning.category
    assert (
        "The distribution A 0.1.0 was required by the input requirement a but ultimately excluded "
        "by configured excludes: a"
    ) == str(warning.message)

    warning = events[1]
    assert PEXWarning == warning.category
    assert (
        "The distribution B 0.1.0 was required by the input requirement b but ultimately excluded "
        "by configured excludes: b"
    ) == str(warning.message)

    pex_builder.freeze()

    assert ["a", "b"] == list(pex_info.requirements)
    assert ["a", "b"] == list(pex_info.excluded)
    assert sorted(("C", "D", "E", "F")) == sorted(pex_info.distributions), (
        "The dependency manager should have excluded the root reqs itself but relied on deep "
        "excludes plumbed to Pip to exclude transitive dependencies of the roots."
    )


def test_exclude_transitive_assert(dist_graph):
    # type: (DistGraph) -> None

    dependency_manager = DependencyManager(
        requirements=OrderedSet(dist_graph.root_reqs), distributions=OrderedSet(dist_graph.dists)
    )

    pex_info = PexInfo.default()
    pex_builder = PEXBuilder(pex_info=pex_info)
    with pytest.raises(AssertionError) as exec_info:
        dependency_manager.configure(
            pex_builder, dependency_configuration=DependencyConfiguration.create(["c"])
        )
    assert reportable_unexpected_error_msg(
        "The deep --exclude mechanism failed to exclude C 0.1.0 from transitive requirements. "
        "It should have been excluded by configured excludes: c but was not."
    ) in str(exec_info.value), str(exec_info.value)

    dependency_manager = DependencyManager(
        requirements=OrderedSet(dist_graph.root_reqs),
        distributions=OrderedSet((dist_graph.dist(name) for name in ("A", "B", "D", "F"))),
    )
    dependency_manager.configure(
        pex_builder, dependency_configuration=DependencyConfiguration.create(["c"])
    )
    pex_builder.freeze()

    assert ["a", "b"] == list(pex_info.requirements)
    assert ["c"] == list(pex_info.excluded)
    expected_dists = [dist_graph.dist(name) for name in ("A", "B", "D", "F")]
    assert {
        os.path.basename(dist.location): dist.fingerprint for dist in expected_dists
    } == pex_info.distributions



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import functools
import os
import sys
from textwrap import dedent

from pex.dist_metadata import Requirement
from pex.enum import Enum
from pex.pep_440 import Version
from pex.targets import LocalInterpreter, Target
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Iterable, Optional, Tuple, Union


@functools.total_ordering
class PipVersionValue(Enum.Value):
    @classmethod
    def overridden(cls):
        # type: () -> Optional[PipVersionValue]
        if not hasattr(cls, "_overridden"):
            setattr(cls, "_overridden", None)

            # We make an affordance for CI with a purposefully undocumented PEX env var.
            overriden_value = os.environ.get("_PEX_PIP_VERSION")
            if overriden_value:
                for version in cls._iter_values():
                    if version.value == overriden_value:
                        setattr(cls, "_overridden", version)
                        break
        return cast("Optional[PipVersionValue]", getattr(cls, "_overridden"))

    def __init__(
        self,
        version,  # type: str
        name=None,  # type: Optional[str]
        requirement=None,  # type: Optional[str]
        setuptools_version=None,  # type: Optional[str]
        wheel_version=None,  # type: Optional[str]
        requires_python=None,  # type: Optional[str]
        hidden=False,  # type: bool
    ):
        # type: (...) -> None
        super(PipVersionValue, self).__init__(name or version)

        def to_requirement(
            project_name,  # type: str
            project_version=None,  # type: Optional[str]
        ):
            # type: (...) -> str
            return (
                "{project_name}=={project_version}".format(
                    project_name=project_name, project_version=project_version
                )
                if project_version
                else project_name
            )

        self.version = Version(version)
        self.requirement = requirement or to_requirement("pip", version)
        self.setuptools_requirement = to_requirement("setuptools", setuptools_version)
        self.wheel_requirement = to_requirement("wheel", wheel_version)
        self.requires_python = SpecifierSet(requires_python) if requires_python else None
        self.hidden = hidden

    @property
    def requirements(self):
        # type: () -> Iterable[str]
        return self.requirement, self.setuptools_requirement, self.wheel_requirement

    def requires_python_applies(self, target=None):
        # type: (Optional[Union[Version,Target]]) -> bool
        if not self.requires_python:
            return True

        if isinstance(target, Version):
            return str(target) in self.requires_python

        return LocalInterpreter.create(
            interpreter=target.get_interpreter() if target else None
        ).requires_python_applies(
            requires_python=self.requires_python,
            source=Requirement.parse(self.requirement),
        )

    def __lt__(self, other):
        if not isinstance(other, PipVersionValue):
            return NotImplemented
        return self.version < other.version


class LatestPipVersion(object):
    def __get__(self, obj, objtype=None):
        if not hasattr(self, "_latest"):
            self._latest = max(
                (version for version in PipVersionValue._iter_values() if not version.hidden),
                key=lambda pv: pv.version,
            )
        return self._latest


class DefaultPipVersion(object):
    def __init__(self, preferred):
        # type: (Iterable[PipVersionValue]) -> None
        self._preferred = preferred

    def __get__(self, obj, objtype=None):
        if not hasattr(self, "_default"):
            self._default = None
            current_version = Version(".".join(map(str, sys.version_info[:3])))
            preferred_versions = (
                [PipVersionValue.overridden()] if PipVersionValue.overridden() else self._preferred
            )
            for preferred_version in preferred_versions:
                if preferred_version.requires_python_applies(current_version):
                    self._default = preferred_version
                    break
            if self._default is None:
                applicable_versions = tuple(
                    version
                    for version in PipVersionValue._iter_values()
                    if not version.hidden and version.requires_python_applies(current_version)
                )
                if not applicable_versions:
                    raise ValueError(
                        dedent(
                            """\
                            No version of Pip supported by Pex works with {python}.
                            The supported Pip versions are:
                            {versions}
                            """
                        ).format(
                            python=sys.executable,
                            versions=", ".join(
                                version.value
                                for version in PipVersionValue._iter_values()
                                if not version.hidden
                            ),
                        )
                    )
                self._default = max(applicable_versions, key=lambda pv: pv.version)
        return self._default


class PipVersion(Enum["PipVersionValue"]):
    @classmethod
    def values(cls):
        # type: () -> Tuple[PipVersionValue, ...]
        if cls._values is None:
            cls._values = tuple(
                version
                for version in PipVersionValue._iter_values()
                if version is PipVersionValue.overridden() or not version.hidden
            )
        return cls._values

    v20_3_4_patched = PipVersionValue(
        name="20.3.4-patched",
        version="20.3.4+patched",
        requirement=(
            "pip @ git+https://github.com/pex-tool/pip@386a54f097ece66775d0c7f34fd29bb596c6b0be"
        ),
        wheel_version="0.37.1",
        requires_python="<3.12",
    )

    # TODO(John Sirois): Expose setuptools and wheel version flags - these don't affect
    #  Pex; so we should allow folks to experiment with upgrade easily:
    #  https://github.com/pex-tool/pex/issues/1895

    v22_2_2 = PipVersionValue(
        version="22.2.2",
        setuptools_version="65.3.0",
        wheel_version="0.37.1",
        requires_python=">=3.7,<3.12",
    )

    v22_3 = PipVersionValue(
        version="22.3",
        setuptools_version="65.5.0",
        wheel_version="0.37.1",
        requires_python=">=3.7,<3.12",
    )

    v22_3_1 = PipVersionValue(
        version="22.3.1",
        setuptools_version="65.5.1",
        wheel_version="0.37.1",
        requires_python=">=3.7,<3.12",
    )

    v23_0 = PipVersionValue(
        version="23.0",
        setuptools_version="67.2.0",
        wheel_version="0.38.4",
        requires_python=">=3.7,<3.12",
    )

    v23_0_1 = PipVersionValue(
        version="23.0.1",
        setuptools_version="67.4.0",
        wheel_version="0.38.4",
        requires_python=">=3.7,<3.12",
    )

    v23_1 = PipVersionValue(
        version="23.1",
        setuptools_version="67.6.1",
        wheel_version="0.40.0",
        requires_python=">=3.7,<3.12",
    )

    v23_1_1 = PipVersionValue(
        version="23.1.1",
        setuptools_version="67.7.1",
        wheel_version="0.40.0",
        requires_python=">=3.7,<3.12",
    )

    v23_1_2 = PipVersionValue(
        version="23.1.2",
        setuptools_version="67.7.2",
        wheel_version="0.40.0",
        requires_python=">=3.7,<3.12",
    )

    v23_2 = PipVersionValue(
        version="23.2",
        setuptools_version="68.0.0",
        wheel_version="0.40.0",
        requires_python=">=3.7,<3.13",
    )

    v23_3_1 = PipVersionValue(
        version="23.3.1",
        # N.B.: The setuptools 68.2.2 release was available on 10/21/2023 (the Pip 23.3.1 release
        # date) but 68.0.0 is the last setuptools version to support 3.7.
        setuptools_version="68.0.0",
        wheel_version="0.41.2",
        requires_python=">=3.7,<3.13",
    )

    v23_3_2 = PipVersionValue(
        version="23.3.2",
        # N.B.: The setuptools 69.0.2 release was available on 12/17/2023 (the Pip 23.3.2 release
        # date) but 68.0.0 is the last setuptools version to support 3.7.
        setuptools_version="68.0.0",
        wheel_version="0.42.0",
        requires_python=">=3.7,<3.13",
    )

    v24_0 = PipVersionValue(
        version="24.0",
        # N.B.: The setuptools 69.0.3 release was available on 2/03/2024 (the Pip 24.0 release
        # date) but 68.0.0 is the last setuptools version to support 3.7.
        setuptools_version="68.0.0",
        wheel_version="0.42.0",
        requires_python=">=3.7,<3.13",
    )

    v24_1 = PipVersionValue(
        version="24.1",
        setuptools_version="70.1.0",
        wheel_version="0.43.0",
        requires_python=">=3.8,<3.14",
    )

    v24_1_1 = PipVersionValue(
        version="24.1.1",
        setuptools_version="70.1.1",
        wheel_version="0.43.0",
        requires_python=">=3.8,<3.14",
    )

    v24_1_2 = PipVersionValue(
        version="24.1.2",
        setuptools_version="70.2.0",
        wheel_version="0.43.0",
        requires_python=">=3.8,<3.14",
    )

    VENDORED = v20_3_4_patched
    LATEST = LatestPipVersion()
    DEFAULT = DefaultPipVersion(preferred=(VENDORED, v23_2, v24_1))



from pex.pep_427 import InstallableType


def get_installable_type_flag(installable_type):
    # type: (InstallableType.Value) -> str
    return (
        "--no-pre-install-wheels"
        if installable_type is InstallableType.WHEEL_FILE
        else "--pre-install-wheels"
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
from argparse import Namespace, _ActionsContainer

from pex.build_system import pep_517
from pex.dependency_configuration import DependencyConfiguration
from pex.dist_metadata import DistMetadata, Requirement
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.interpreter import PythonInterpreter
from pex.jobs import Raise, SpawnedJob, execute_parallel
from pex.pep_427 import InstallableType
from pex.pip.version import PipVersionValue
from pex.resolve.configured_resolve import resolve
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolver_configuration import PipConfiguration
from pex.resolve.resolvers import Resolver, Untranslatable
from pex.targets import LocalInterpreter, Target, Targets
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator, List, Optional, Set, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class BuiltProject(object):
    target = attr.ib()  # type: Target
    fingerprinted_distribution = attr.ib()  # type: FingerprintedDistribution

    def as_requirement(self):
        # type: () -> Requirement
        return self.fingerprinted_distribution.distribution.as_requirement()

    @property
    def requires_dists(self):
        # type: () -> Tuple[Requirement, ...]
        return self.fingerprinted_distribution.distribution.metadata.requires_dists


@attr.s(frozen=True)
class Projects(object):
    paths = attr.ib(default=())  # type: Tuple[str, ...]

    @paths.validator
    def _is_project_dir(
        self,
        attribute,  # type: attr.Attribute
        value,  # type: Tuple[str, ...]
    ):
        # type: (...) -> None
        invalid_paths = []  # type: List[str]
        for path in value:
            if os.path.isfile(os.path.join(path, "pyproject.toml")):
                continue
            if os.path.isfile(os.path.join(path, "setup.py")):
                continue
            invalid_paths.append(path)

        if invalid_paths:
            raise ValueError(
                "The following --project paths do not appear to point to directories containing "
                "Python projects:\n"
                "{invalid_paths}".format(
                    invalid_paths="\n".join(
                        "{index}. {path}".format(index=index, path=path)
                        for index, path in enumerate(invalid_paths, start=1)
                    )
                )
            )

    def build(
        self,
        targets,  # type: Targets
        pip_configuration,  # type: PipConfiguration
        compile_pyc=False,  # type: bool
        ignore_errors=False,  # type: bool
        result_type=InstallableType.INSTALLED_WHEEL_CHROOT,  # type: InstallableType.Value
        dependency_config=DependencyConfiguration(),  # type: DependencyConfiguration
    ):
        # type: (...) -> Iterator[BuiltProject]

        resolve_result = resolve(
            targets=targets,
            requirement_configuration=RequirementConfiguration(requirements=self.paths),
            resolver_configuration=attr.evolve(pip_configuration, transitive=False),
            compile_pyc=compile_pyc,
            ignore_errors=ignore_errors,
            result_type=result_type,
            dependency_configuration=dependency_config,
        )
        for resolved_distribution in resolve_result.distributions:
            yield BuiltProject(
                target=resolved_distribution.target,
                fingerprinted_distribution=resolved_distribution.fingerprinted_distribution,
            )

    def collect_requirements(
        self,
        resolver,  # type: Resolver
        interpreter=None,  # type: Optional[PythonInterpreter]
        pip_version=None,  # type: Optional[PipVersionValue]
        max_jobs=None,  # type: Optional[int]
    ):
        # type: (...) -> Iterator[Requirement]

        target = LocalInterpreter.create(interpreter)

        def spawn_func(project_directory):
            # type: (str) -> SpawnedJob[DistMetadata]
            return pep_517.spawn_prepare_metadata(
                project_directory, target, resolver, pip_version=pip_version
            )

        seen = set()  # type: Set[Requirement]
        for dist_metadata in execute_parallel(
            self.paths,
            spawn_func=spawn_func,
            error_handler=Raise[str, DistMetadata](Untranslatable),
            max_jobs=max_jobs,
        ):
            for req in dist_metadata.requires_dists:
                if req not in seen:
                    seen.add(req)
                    yield req

    def __len__(self):
        # type: () -> int
        return len(self.paths)


def register_options(
    parser,  # type: _ActionsContainer
    help,  # type: str
):
    # type: (...) -> None

    parser.add_argument(
        "--project",
        dest="projects",
        metavar="DIR",
        default=[],
        type=str,
        action="append",
        help=help,
    )


def get_projects(options):
    # type: (Namespace) -> Projects
    return Projects(paths=tuple(options.projects))



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import json
import logging
import os
import re
import signal
import subprocess
import time

from pex.atomic_directory import atomic_directory
from pex.common import safe_open, safe_rmtree
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraint
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InvalidVirtualenvError, Virtualenv
from testing import PEX_TEST_DEV_ROOT

if TYPE_CHECKING:
    from typing import List, Optional, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


logger = logging.getLogger(__name__)


DEVPI_DIR = os.path.join(PEX_TEST_DEV_ROOT, "devpi")


@attr.s(frozen=True)
class Pidfile(object):
    _PATH = os.path.join(DEVPI_DIR, "pidfile")

    @classmethod
    def load(cls):
        # type: () -> Optional[Pidfile]
        try:
            with open(cls._PATH) as fp:
                data = json.load(fp)
            return cls(url=data["url"], pid=data["pid"])
        except (OSError, IOError, ValueError, KeyError) as e:
            logger.warning(
                "Failed to load devpi-server pid file from {path}: {err}".format(
                    path=cls._PATH, err=e
                )
            )
            return None

    @staticmethod
    def _read_base_url(
        log,  # type: str
        timeout,  # type: float
    ):
        # type: (...) -> Optional[str]
        start = time.time()
        while time.time() - start < timeout:
            with open(log) as fp:
                for line in fp:
                    if line.endswith(os.linesep):
                        match = re.search(r"Serving on (?P<url>http://\S+)$", line)
                        if match:
                            return match.group("url")
        return None

    @classmethod
    def record(
        cls,
        log,  # type: str
        pid,  # type: int
        timeout=5.0,  # type: float
    ):
        # type: (...) -> Optional[Pidfile]
        base_url = cls._read_base_url(log, timeout)
        if not base_url:
            return None

        url = "{base_url}/root/pypi/+simple/".format(base_url=base_url)
        with safe_open(cls._PATH, "w") as fp:
            json.dump(dict(url=url, pid=pid), fp, indent=2, sort_keys=True)
        return cls(url=url, pid=pid)

    url = attr.ib()  # type: str
    pid = attr.ib()  # type: int

    def alive(self):
        # type: () -> bool
        # TODO(John Sirois): Handle pid rollover
        try:
            os.kill(self.pid, 0)
            return True
        except OSError as e:
            if e.errno == errno.ESRCH:  # No such process.
                return False
            raise

    def kill(self):
        # type: () -> None
        os.kill(self.pid, signal.SIGTERM)


@attr.s(frozen=True)
class DevpiServer(object):
    python = attr.ib()  # type: str
    script = attr.ib()  # type: str
    serverdir = attr.ib()  # type: str

    def launch_args(self, *extra_args):
        # type: (*str) -> List[str]
        return [self.python, self.script, "--serverdir", self.serverdir] + list(extra_args)


def ensure_devpi_server():
    # type: () -> DevpiServer

    venv_dir = os.path.join(DEVPI_DIR, "venv")
    try:
        venv = Virtualenv(venv_dir=venv_dir)
    except InvalidVirtualenvError as e:
        logger.warning(str(e))
        safe_rmtree(venv_dir)
        with atomic_directory(venv_dir) as atomic_venvdir:
            if not atomic_venvdir.is_finalized():
                logger.info("Installing devpi-server...")
                lock = os.path.join(os.path.dirname(__file__), "devpi-server.lock")
                python = PythonInterpreter.latest_release_of_min_compatible_version(
                    InterpreterConstraint.parse(">=3.8,<3.13").iter_matching()
                )
                Virtualenv.create_atomic(venv_dir=atomic_venvdir, interpreter=python, force=True)
                subprocess.check_call(
                    args=[
                        python.binary,
                        "-m",
                        "pex.cli",
                        "venv",
                        "create",
                        "--lock",
                        lock,
                        "-d",
                        atomic_venvdir.work_dir,
                    ]
                )
        venv = Virtualenv(venv_dir=venv_dir)

    serverdir = os.path.join(DEVPI_DIR, "serverdir")
    with atomic_directory(serverdir) as atomic_serverdir:
        if not atomic_serverdir.is_finalized():
            logger.info("Initializing devpi-server...")
            logger.info("Using {}".format(venv.bin_path("devpi-init")))
            subprocess.check_call(
                args=[
                    venv.interpreter.binary,
                    venv.bin_path("devpi-init"),
                    "--serverdir",
                    atomic_serverdir.work_dir,
                ]
            )

    return DevpiServer(
        python=venv.interpreter.binary, script=venv.bin_path("devpi-server"), serverdir=serverdir
    )


@attr.s(frozen=True)
class LaunchResult(object):
    url = attr.ib()  # type: str
    already_running = attr.ib()  # type: bool


def launch(
    host,  # type: str
    port,  # type: int
    timeout,  # type: float
    max_connection_retries,  # type: int
    request_timeout,  # type: int
):
    # type: (...) -> Union[str, LaunchResult]

    pidfile = Pidfile.load()
    if pidfile and pidfile.alive():
        return LaunchResult(url=pidfile.url, already_running=True)

    devpi_server = ensure_devpi_server()

    # Not proper daemonization, but good enough.
    log = os.path.join(DEVPI_DIR, "log.txt")
    with safe_open(log, "w") as fp:
        process = subprocess.Popen(
            args=devpi_server.launch_args(
                "--host",
                host,
                "--port",
                str(port),
                "--replica-max-retries",
                str(max_connection_retries),
                "--request-timeout",
                str(request_timeout),
            ),
            cwd=DEVPI_DIR,
            preexec_fn=os.setsid,
            stdout=fp.fileno(),
            stderr=subprocess.STDOUT,
        )

    pidfile = Pidfile.record(log=log, pid=process.pid, timeout=timeout)
    if not pidfile:
        try:
            os.kill(process.pid, signal.SIGKILL)
        except OSError as e:
            if e.errno != errno.ESRCH:  # No such process.
                raise
        return log

    return LaunchResult(url=pidfile.url, already_running=False)


def shutdown():
    # type: () -> bool

    pidfile = Pidfile.load()
    if not pidfile:
        return False

    logger.info("Killing devpi server {url} @ {pid}".format(url=pidfile.url, pid=pidfile.pid))
    pidfile.kill()
    return True



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import filecmp
import os.path
import re
import shutil
import subprocess
import sys
from textwrap import dedent

import pytest

from pex.atomic_directory import atomic_directory
from pex.dist_metadata import find_distribution
from pex.interpreter import PythonInterpreter
from pex.pep_425 import CompatibilityTags
from pex.pep_427 import InstallableType
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.locked_resolve import (
    Artifact,
    FileArtifact,
    LockedRequirement,
    LockedResolve,
    LockStyle,
)
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.resolve.resolved_requirement import Pin
from pex.sorted_tuple import SortedTuple
from pex.third_party.packaging.tags import Tag
from pex.typing import TYPE_CHECKING, cast
from pex.venv.virtualenv import InstallationChoice, Virtualenv
from testing import (
    IS_PYPY,
    IS_X86_64,
    PY39,
    PY310,
    PY_VER,
    IntegResults,
    ensure_python_interpreter,
    make_env,
    re_exact,
)
from testing.cli import run_pex3
from testing.find_links import FindLinksRepo

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Mapping, Optional, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class SessionFixtures(object):
    find_links = attr.ib()  # type: str
    initial_lock = attr.ib()  # type: str


@pytest.fixture(scope="session")
def session_fixtures(shared_integration_test_tmpdir):
    # type: (str) -> SessionFixtures

    test_lock_sync_chroot = os.path.join(shared_integration_test_tmpdir, "test_lock_sync_chroot")
    with atomic_directory(test_lock_sync_chroot) as chroot:
        if not chroot.is_finalized():
            pip_version = PipVersion.DEFAULT
            find_links = os.path.join(chroot.work_dir, "find_links")
            find_links_repo = FindLinksRepo.create(find_links, pip_version)

            def host_requirements(*requirements):
                # type: (*str) -> None
                result = find_links_repo.resolver.resolve_requirements(
                    requirements,
                    result_type=InstallableType.WHEEL_FILE,
                )
                for resolved_distribution in result.distributions:
                    find_links_repo.host(resolved_distribution.distribution.location)

            # N.B.: Since we are setting up a find links repo for offline lock resolves, we need to
            # grab at least one distribution online to allow the current Pip version to bootstrap
            # itself if needed.
            host_requirements(
                "cowsay==5.0.0",
                pip_version.setuptools_requirement,
                pip_version.wheel_requirement,
            )
            find_links_repo.make_sdist("spam", version="1")
            find_links_repo.make_wheel("spam", version="1")
            for project_name in "foo", "bar", "baz":
                find_links_repo.make_sdist(project_name, version="1", install_reqs=["spam"])
                find_links_repo.make_wheel(project_name, version="1", install_reqs=["spam"])

            run_pex3(
                "lock",
                "create",
                "--path-mapping",
                "FL|{find_links}|Find Links Repo".format(find_links=find_links),
                "--no-pypi",
                "-f",
                find_links,
                "cowsay",
                "foo",
                "bar",
                "-o",
                os.path.join(chroot.work_dir, "lock.json"),
                "--indent",
                "2",
            ).assert_success()

            host_requirements("cowsay<6.1")
            find_links_repo.make_sdist("spam", version="2", install_reqs=["foo"])
            find_links_repo.make_wheel("spam", version="2", install_reqs=["foo"])
            for project_name in "foo", "bar", "baz":
                find_links_repo.make_sdist(project_name, version="2")
                find_links_repo.make_wheel(project_name, version="2")

    return SessionFixtures(
        find_links=os.path.join(test_lock_sync_chroot, "find_links"),
        initial_lock=os.path.join(test_lock_sync_chroot, "lock.json"),
    )


@pytest.fixture(scope="session")
def find_links(session_fixtures):
    # type: (SessionFixtures) -> str
    return session_fixtures.find_links


@pytest.fixture(scope="session")
def path_mappings(find_links):
    # type: (str) -> PathMappings
    return PathMappings((PathMapping(path=find_links, name="FL"),))


@pytest.fixture
def initial_lock(
    tmpdir,  # type: str
    session_fixtures,  # type: SessionFixtures
):
    # type: (...) -> str
    initial_lock_copy = os.path.join(str(tmpdir), "lock.json")
    shutil.copy(session_fixtures.initial_lock, initial_lock_copy)
    return initial_lock_copy


@pytest.fixture
def repo_args(find_links):
    # type: (str) -> List[str]
    return ["--no-index", "-f", find_links]


@pytest.fixture
def path_mapping_args(path_mappings):
    # type: (PathMappings) -> List[str]
    args = []
    for path_mapping in path_mappings.mappings:
        args.append("--path-mapping")
        args.append(path_mapping.name + "|" + path_mapping.path)
    return args


NO_OUTPUT = r"^$"


def run_sync(
    *args,  # type: str
    **popen_kwargs  # type: Any
):
    # type: (...) -> IntegResults
    return run_pex3("lock", "sync", *args, **popen_kwargs)


def pin(
    project_name,  # type: str
    version,  # type: str
):
    # type: (...) -> Pin
    return Pin(ProjectName(project_name), Version(version))


def assert_lock(
    lock,  # type: Union[str, Lockfile]
    path_mappings,  # type: PathMappings
    expected_pins,  # type: Iterable[Pin]
):
    # type: (...) -> LockedResolve

    lock_file = (
        lock if isinstance(lock, Lockfile) else json_codec.load(lock, path_mappings=path_mappings)
    )
    assert len(lock_file.locked_resolves) == 1
    locked_resolve = lock_file.locked_resolves[0]
    assert sorted(expected_pins, key=str) == sorted(
        (locked_requirement.pin for locked_requirement in locked_resolve.locked_requirements),
        key=str,
    )
    return locked_resolve


def assert_venv(
    venv,  # type: Union[str, Virtualenv]
    expected_pins,  # type: Iterable[Pin]
):
    # type: (...) -> Virtualenv
    virtualenv = venv if isinstance(venv, Virtualenv) else Virtualenv(venv_dir=venv)
    assert sorted(expected_pins, key=str) == sorted(
        (
            Pin(distribution.metadata.project_name, distribution.metadata.version)
            for distribution in virtualenv.iter_distributions(rescan=True)
        ),
        key=str,
    )
    return virtualenv


def assert_lock_matches_venv(
    lock,  # type: str
    path_mappings,  # type: PathMappings
    venv,  # type: Union[str, Virtualenv]
    expected_pins,  # type: Iterable[Pin]
):
    # type: (...) -> Lockfile

    lock_file = json_codec.load(lock, path_mappings=path_mappings)
    assert_lock(lock_file, path_mappings, expected_pins)
    assert_venv(venv, expected_pins)
    return lock_file


def test_sync_implicit_create(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_sync("cowsay==5.0", "--lock", lock, *repo_args).assert_success()
    assert_lock(lock, path_mappings, expected_pins=[pin("cowsay", "5.0")])


def test_sync_implicit_create_lock_create_equivalence(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock1 = os.path.join(str(tmpdir), "lock1.json")
    run_sync("cowsay==5.0", "--lock", lock1, *repo_args).assert_success()

    lock2 = os.path.join(str(tmpdir), "lock2.json")
    run_pex3("lock", "create", "cowsay==5.0", "-o", lock2, *repo_args).assert_success()
    assert filecmp.cmp(lock1, lock2, shallow=False)


def test_sync_implicit_create_venv(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    venv_dir = os.path.join(str(tmpdir), "venv")
    run_sync("cowsay==5.0", "--lock", lock, "--venv", venv_dir, *repo_args).assert_success()
    venv = Virtualenv(venv_dir)
    assert_lock_matches_venv(
        lock=lock, path_mappings=path_mappings, venv=venv, expected_pins=[pin("cowsay", "5.0")]
    )
    assert b"| Moo! |" in subprocess.check_output(args=[venv.bin_path("cowsay"), "Moo!"])


def test_sync_implicit_lock_create_venv_create_run(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    venv_dir = os.path.join(str(tmpdir), "venv")
    run_sync(
        *(repo_args + ["cowsay==5.0", "--lock", lock, "--venv", venv_dir, "--", "cowsay", "Moo!"])
    ).assert_success(expected_output_re=r".*\| Moo! \|.*", re_flags=re.DOTALL)
    assert_lock_matches_venv(
        lock=lock, path_mappings=path_mappings, venv=venv_dir, expected_pins=[pin("cowsay", "5.0")]
    )


def test_sync_noop(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None
    locked_resolve = assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )
    run_sync(
        "cowsay", "foo", "bar", "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            "No updates for lock generated by {platform}.".format(
                platform=locked_resolve.platform_tag
            )
        ),
    )
    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )


def test_sync_update(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None
    locked_resolve = assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )

    run_sync(
        "cowsay", "foo>1", "bar", "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Updated foo from 1 to 2
                Updates to lock input requirements:
                  Updated 'foo' to 'foo>1'
                """
            ).format(platform=locked_resolve.platform_tag)
        ),
    )
    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "2"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )


def test_sync_add(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None
    locked_resolve = assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )

    run_sync(
        "cowsay", "foo", "bar", "baz", "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Added baz 2
                Updates to lock input requirements:
                  Added 'baz'
                """
            ).format(platform=locked_resolve.platform_tag)
        ),
    )
    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("baz", "2"),
            pin("spam", "1"),
        ],
    )


def test_sync_remove(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None
    locked_resolve = assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )

    run_sync(
        "cowsay", "foo", "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Deleted bar 1
                Updates to lock input requirements:
                  Deleted 'bar'
                """
            ).format(platform=locked_resolve.platform_tag)
        ),
    )
    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[pin("cowsay", "5.0"), pin("foo", "1"), pin("spam", "1")],
    )


def test_sync_complex(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None
    locked_resolve = assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )

    # N.B.: Update foo, remove bar, add baz.
    run_sync(
        "cowsay", "foo<3", "baz<2", "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Deleted bar 1
                  Updated foo from 1 to 2
                  Added baz 1
                Updates to lock input requirements:
                  Deleted 'bar'
                  Updated 'foo' to 'foo<3'
                  Added 'baz<2'
                """
            ).format(platform=locked_resolve.platform_tag)
        ),
    )
    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "2"),
            pin("baz", "1"),
            pin("spam", "1"),
        ],
    )


def test_sync_configuration(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> None

    lock_file = json_codec.load(initial_lock, path_mappings=path_mappings)
    assert_lock(
        lock_file,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )
    assert lock_file.style is LockStyle.STRICT

    run_sync(
        "cowsay",
        "foo>1",
        "bar",
        "--style",
        "sources",
        "--lock",
        initial_lock,
        *(repo_args + path_mapping_args)
    ).assert_success(
        expected_output_re=NO_OUTPUT,
        expected_error_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Updated bar 1 artifacts:
                    + file://${{FL}}/bar-1.tar.gz
                  Updated foo from 1 to 2
                  Updated spam 1 artifacts:
                    + file://${{FL}}/spam-1.tar.gz
                Updates to lock input requirements:
                  Updated 'foo' to 'foo>1'
                """
            ).format(platform=lock_file.locked_resolves[0].platform_tag)
        ),
    )
    lock_file = json_codec.load(initial_lock, path_mappings=path_mappings)
    assert_lock(
        lock_file,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "2"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )
    assert lock_file.style is LockStyle.SOURCES


@pytest.fixture
def initial_venv(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
):
    # type: (...) -> str
    venv_dir = os.path.join(str(tmpdir), "venv")
    run_pex3(
        "venv", "create", "-d", venv_dir, "--lock", initial_lock, *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=venv_dir,
        expected_pins=[pin("cowsay", "5.0"), pin("foo", "1"), pin("bar", "1"), pin("spam", "1")],
    )
    return venv_dir


def test_sync_venv_noop(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    # N.B.: There are no changed requirements and we don't pass "--yes", which would cause a
    # blocking input read if there needed to be any venv distribution deletes.
    run_sync(
        "cowsay",
        "foo",
        "bar",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )


def test_sync_venv_update(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    # N.B.: The only changed requirement is "bar" -> "bar==2"
    run_sync(
        "--yes",
        "cowsay",
        "foo",
        "bar==2",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "2"),
            pin("spam", "1"),
        ],
    )


def test_sync_venv_add(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    # N.B.: The only changed requirement is adding "baz" and we don't pass "--yes", which would be
    # required if there were venv dist deletes.
    run_sync(
        "cowsay",
        "foo",
        "bar",
        "baz",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("baz", "2"),
            pin("spam", "1"),
        ],
    )


def test_sync_venv_remove(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    # N.B.: The only changed requirement is "cowsay" being removed.
    run_sync(
        "--yes",
        "foo",
        "bar",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[pin("foo", "1"), pin("bar", "1"), pin("spam", "1")],
    )


def test_sync_venv_complex(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    # N.B.: The "cowsay" and "foo" requirements are removed, "bar" -> "bar>1" and "baz" is added.
    # Since bar and baz promote to v2, there is no longer any transitive dependency on spam
    # and so it is implicitly removed.
    run_sync(
        "--yes",
        "bar>1",
        "baz",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[pin("bar", "2"), pin("baz", "2")],
    )


def test_sync_venv_transitive_to_direct_and_vice_versa(
    repo_args,  # type: List[str]
    initial_lock,  # type: str
    path_mappings,  # type: PathMappings
    path_mapping_args,  # type: List[str]
    initial_venv,  # type: str
):
    # type: (...) -> None

    assert_lock(
        initial_lock,
        path_mappings,
        expected_pins=[
            pin("cowsay", "5.0"),
            pin("foo", "1"),
            pin("bar", "1"),
            pin("spam", "1"),
        ],
    )

    run_sync(
        "--yes",
        "spam",
        "--lock",
        initial_lock,
        "--venv",
        initial_venv,
        *(repo_args + path_mapping_args)
    ).assert_success()

    # N.B.: The spam project was locked at version 1 in the initial lock, but we ask for it here as
    # a new top-level requirement unconstrained; so we expect the latest version to be resolved.
    # N.B.: Since foo was in the initial lock at version 1, it should stay undisturbed at 1 even
    # though foo is a transitive dependency of spam and foo 2 is available.
    assert_lock_matches_venv(
        lock=initial_lock,
        path_mappings=path_mappings,
        venv=initial_venv,
        expected_pins=[pin("spam", "2"), pin("foo", "1")],
    )


skip_cowsay6_for_python27 = pytest.mark.skipif(
    sys.version_info[0] < 3,
    reason=(
        "The cowsay 6.0 distribution is mistakenly resolvable by Python 2.7 (it does not have "
        "Requires-Python metadata), but it uses Python 3 syntax"
    ),
)


def assert_cowsay5(venv):
    # type: (Virtualenv) -> None
    assert b"| Moo! |" in subprocess.check_output(args=[venv.bin_path("cowsay"), "Moo!"])


@skip_cowsay6_for_python27
def test_sync_venv_run(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    # N.B.: The cowsay 5.0 -> cowsay 6.0 transition is picked, in part, because cowsay 5.0 is sdist
    # only on PyPI. Older versions of Pip (those that come with the venvs created by Python 3.10
    # and older), install sdists as `.egg-info` distributions in site-packages instead of
    # regularizing to `.dist-info` as is done in newer versions of Pip. In that way our CI test
    # matrix ensures we test both `.dist-info` and `.egg-info` distributions are properly detected
    # and removed when appropriate.

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3("lock", "create", "cowsay==5.0", "-o", lock, *repo_args).assert_success()
    # N.B.: There is no Pip in the lock.
    assert_lock(lock, path_mappings, expected_pins=[pin("cowsay", "5.0")])

    venv_dir = os.path.join(str(tmpdir), "venv")
    run_pex3("venv", "create", "-d", venv_dir, "--lock", lock, "--pip", *repo_args).assert_success()
    venv = Virtualenv(venv_dir)
    assert (
        any(ProjectName("pip") == dist.metadata.project_name for dist in venv.iter_distributions())
        > 0
    ), "We expect the initial venv to include Pip."
    assert_cowsay5(venv)

    result = run_sync(
        *(
            repo_args
            + [
                "--yes",
                "cowsay<6.1",
                "--lock",
                lock,
                "--",
                venv.bin_path("cowsay"),
                "-t",
                "A New Moo!",
            ]
        )
    )
    result.assert_success(expected_output_re=r".*\| A New Moo! \|.*", re_flags=re.DOTALL)

    # N.B.: Since the venv now matches the lock, this means Pip and its dist dependencies were
    # nuked, confirming the default --no-retain-pip mode.
    lockfile = assert_lock_matches_venv(
        lock=lock, path_mappings=path_mappings, venv=venv_dir, expected_pins=[pin("cowsay", "6.0")]
    )
    assert (
        dedent(
            """\
            Updates for lock generated by {platform}:
              Updated cowsay from 5 to 6
            Updates to lock input requirements:
              Updated 'cowsay==5.0' to 'cowsay<6.1'
            """
        ).format(platform=lockfile.locked_resolves[0].platform_tag)
        == result.error
    )


def test_sync_venv_dry_run_create(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir)
    run_sync(
        *(
            repo_args
            + [
                "cowsay<6.1",
                "--style",
                "universal",
                "--lock",
                lock,
                "--dry-run",
                "--venv",
                venv_dir,
                "--",
                "cowsay",
                "-t",
                "I would have mooed!",
            ]
        )
    ).assert_success(
        expected_output_re=re_exact(
            dedent(
                """\
                Would lock 1 project for platform universal:
                  cowsay 6
                Would sync venv at {venv_dir} and run the following command in it:
                  {cowsay} -t 'I would have mooed!'
                """
            ).format(
                venv_dir=venv_dir,
                cowsay=venv.bin_path("cowsay"),
            )
        ),
        expected_error_re=NO_OUTPUT,
    )


def test_sync_venv_dry_run_update(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3("lock", "create", "cowsay==5.0", "-o", lock, *repo_args).assert_success()
    locked_resolve = assert_lock(lock, path_mappings, expected_pins=[pin("cowsay", "5.0")])

    venv_dir = os.path.join(str(tmpdir), "venv")
    run_pex3("venv", "create", "-d", venv_dir, "--lock", lock, "--pip", *repo_args).assert_success()
    venv = Virtualenv(venv_dir)
    assert_cowsay5(venv)

    run_sync(
        *(
            repo_args
            + [
                "cowsay<6.1",
                "--lock",
                lock,
                "--dry-run",
                "--",
                "cowsay",
                "-t",
                "I would have mooed!",
            ]
        ),
        # Simulate an activated venv with its bin dir inserted n the PATH.
        env=make_env(
            PATH=os.pathsep.join(
                [venv.bin_dir] + os.environ.get("PATH", os.defpath).split(os.pathsep)
            )
        )
    ).assert_success(
        expected_output_re=re_exact(
            dedent(
                """\
                Updates for lock generated by {platform}:
                  Would update cowsay from 5 to 6
                Updates to lock input requirements:
                  Would update 'cowsay==5.0' to 'cowsay<6.1'
                Would sync venv at {venv_dir} and run the following command in it:
                  {cowsay} -t 'I would have mooed!'
                """
            ).format(
                platform=locked_resolve.platform_tag,
                venv_dir=venv_dir,
                cowsay=venv.bin_path("cowsay"),
            )
        ),
        expected_error_re=NO_OUTPUT,
    )


def test_sync_venv_run_no_retain_pip_preinstalled(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir, install_pip=InstallationChoice.YES)
    pip = find_distribution("pip", search_path=venv.sys_path)
    assert pip is not None

    subprocess.check_call(args=[venv.bin_path("pip"), "install", "cowsay==5.0"] + repo_args)
    assert_cowsay5(venv)

    lock = os.path.join(str(tmpdir), "lock.json")
    run_sync(
        *(
            repo_args
            + [
                "--no-retain-pip",
                "--yes",
                "cowsay==5.0",
                "--lock",
                lock,
                "--",
                venv.bin_path("cowsay"),
                "Moo Two!",
            ]
        )
    ).assert_success(expected_output_re=r".*\| Moo Two! \|.*", re_flags=re.DOTALL)
    assert_lock_matches_venv(
        lock=lock, venv=venv, path_mappings=path_mappings, expected_pins=[pin("cowsay", "5.0")]
    )


@skip_cowsay6_for_python27
def test_sync_venv_run_retain_pip_preinstalled(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir, install_pip=InstallationChoice.YES)
    pip = find_distribution("pip", search_path=venv.sys_path)
    assert pip is not None
    pip_pin = pin("pip", pip.version)

    subprocess.check_call(args=[venv.bin_path("pip"), "install", "cowsay==5.0"] + repo_args)
    assert_cowsay5(venv)

    lock = os.path.join(str(tmpdir), "lock.json")
    run_sync(
        *(
            repo_args
            + [
                "--retain-pip",
                "--yes",
                "cowsay<6.1",
                "--lock",
                lock,
                "--",
                venv.bin_path("cowsay"),
                "-t",
                "A New Moo!",
            ]
        )
    ).assert_success(expected_output_re=r".*\| A New Moo! \|.*", re_flags=re.DOTALL)

    assert_lock(lock, path_mappings, expected_pins=[pin("cowsay", "6.0")])
    assert_venv(venv_dir, expected_pins=[pin("cowsay", "6.0"), pip_pin])

    # And check Pip still works.
    subprocess.check_call(args=[venv.bin_path("pip"), "uninstall", "--yes", "cowsay"])
    assert_venv(venv_dir, expected_pins=[pip_pin])
    subprocess.check_call(args=[venv.bin_path("pip"), "install", "cowsay==5.0"] + repo_args)
    assert_venv(venv_dir, expected_pins=[pin("cowsay", "5.0"), pip_pin])
    assert_cowsay5(venv)


def test_sync_venv_run_retain_pip_no_pip_preinstalled(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
):
    # type: (...) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_sync(
        *(
            repo_args
            + [
                "--no-pip",
                "cowsay==5.0",
                "--lock",
                lock,
                "--venv",
                venv_dir,
                "--",
                "cowsay",
                "Moo!",
            ]
        )
    ).assert_success(expected_output_re=r".*\| Moo! \|.*", re_flags=re.DOTALL)
    lock_file = assert_lock_matches_venv(
        lock=lock, path_mappings=path_mappings, venv=venv_dir, expected_pins=[pin("cowsay", "5.0")]
    )

    venv = Virtualenv(venv_dir)
    run_sync(
        *(
            repo_args
            + [
                "--pip",
                "cowsay==5.0",
                "--lock",
                lock,
                "--",
                venv.bin_path("cowsay"),
                "Moo Two!",
            ]
        )
    ).assert_success(
        expected_output_re=r".*\| Moo Two! \|.*",
        expected_error_re=r".*No updates for lock generated by {platform}\..*".format(
            platform=re.escape(str(lock_file.locked_resolves[0].platform_tag))
        ),
        re_flags=re.DOTALL,
    )
    assert_lock(lock, path_mappings, expected_pins=[pin("cowsay", "5.0")])

    pip = find_distribution("pip", search_path=venv.sys_path, rescan=True)
    assert pip is not None


@skip_cowsay6_for_python27
@pytest.mark.parametrize(
    "retain_pip_args",
    [
        pytest.param([], id="default"),
        pytest.param(["--retain-pip"], id="--retain-pip"),
        pytest.param(["--no-retain-pip"], id="--no-retain-pip"),
    ],
)
def test_sync_venv_run_retain_user_pip(
    tmpdir,  # type: Any
    repo_args,  # type: List[str]
    path_mappings,  # type: PathMappings
    retain_pip_args,  # type: List[str]
):
    # type: (...) -> None

    venv_dir = os.path.join(str(tmpdir), "venv")
    venv = Virtualenv.create(venv_dir, install_pip=InstallationChoice.YES)
    original_pip = find_distribution("pip", search_path=venv.sys_path)
    assert original_pip is not None

    subprocess.check_call(args=[venv.bin_path("pip"), "install", "cowsay==5.0"] + repo_args)
    assert_cowsay5(venv)

    requirements = os.path.join(str(tmpdir), "requirements.txt")
    with open(requirements, "w") as fp:
        print("cowsay<6.1", file=fp)

    constraints = os.path.join(str(tmpdir), "constraints.txt")
    with open(constraints, "w") as fp:
        print(
            "pip!={original_pip_version}".format(original_pip_version=original_pip.version), file=fp
        )

    lock = os.path.join(str(tmpdir), "lock.json")
    run_sync(
        *(
            repo_args
            + retain_pip_args
            + [
                "--pypi",  # N.B.: We need to turn PyPI back on to get at the user Pip.
                "--yes",
                "-r",
                requirements,
                "--constraints",
                constraints,
                "pip",
                "--lock",
                lock,
                "--",
                venv.bin_path("cowsay"),
                "-t",
                "A New Moo!",
            ]
        )
    ).assert_success(expected_output_re=r".*\| A New Moo! \|.*", re_flags=re.DOTALL)

    user_pip = find_distribution("pip", search_path=venv.sys_path, rescan=True)
    assert user_pip is not None
    assert user_pip.version != original_pip.version
    user_pip_pin = pin("pip", user_pip.version)

    assert_lock_matches_venv(
        lock=lock,
        path_mappings=path_mappings,
        venv=venv,
        expected_pins=[pin("cowsay", "6.0"), user_pip_pin],
    )

    # And check Pip still works.
    subprocess.check_call(args=[venv.bin_path("pip"), "uninstall", "--yes", "cowsay"])
    assert_venv(venv_dir, expected_pins=[user_pip_pin])
    subprocess.check_call(args=[venv.bin_path("pip"), "install", "cowsay==5.0"] + repo_args)
    assert_venv(venv_dir, expected_pins=[pin("cowsay", "5.0"), user_pip_pin])
    assert_cowsay5(venv)


def wheel_tag(artifact):
    # type: (Artifact) -> Optional[Tag]
    if not isinstance(artifact, FileArtifact) or artifact.is_source:
        return None
    # Just 1 tag will do from the manylinux compressed tag sets.
    return CompatibilityTags.from_wheel(artifact.filename)[0]


def assert_wheel_tag(
    artifact,  # type: Artifact
    expected_python_tag=None,  # type: Optional[str]
    expected_abi_tag=None,  # type: Optional[str]
    expected_platform_tag=None,  # type: Optional[str]
):
    # type: (...) -> Optional[Tag]
    tag = wheel_tag(artifact)
    if tag is not None:
        if expected_python_tag:
            assert expected_python_tag == tag.interpreter
        if expected_abi_tag:
            assert expected_abi_tag == tag.abi
        if expected_platform_tag:
            assert expected_platform_tag == tag.platform
    return tag


@attr.s(frozen=True)
class LockAnalysis(object):
    pin = attr.ib()  # type: Pin
    artifacts_by_tag = attr.ib()  # type: Mapping[Optional[Tag], Artifact]


def assert_p537_lock(
    lock,  # type: str
    expected_style,  # type: LockStyle.Value
    expected_requires_python=None,  # type: Optional[str]
    expected_python_tag=None,  # type: Optional[str]
    expected_abi_tag=None,  # type: Optional[str]
    expected_platform_tag=None,  # type: Optional[str]
):
    # type: (...) -> LockAnalysis

    lock_file = json_codec.load(lock)
    assert lock_file.style is expected_style
    assert (
        SortedTuple([expected_requires_python])
        if expected_requires_python
        else SortedTuple() == lock_file.requires_python
    )
    assert len(lock_file.locked_resolves) == 1
    locked_resolve = lock_file.locked_resolves[0]
    assert len(locked_resolve.locked_requirements) == 1
    locked_requirement = locked_resolve.locked_requirements[0]
    assert ProjectName("p537") == locked_requirement.pin.project_name

    return LockAnalysis(
        pin=locked_requirement.pin,
        artifacts_by_tag={
            assert_wheel_tag(
                artifact,
                expected_python_tag=expected_python_tag,
                expected_abi_tag=expected_abi_tag,
                expected_platform_tag=expected_platform_tag,
            ): artifact
            for artifact in locked_requirement.iter_artifacts()
        },
    )


skip_unless_p537_compatible = pytest.mark.skipif(
    PY_VER < (3, 6) or PY_VER >= (3, 13) or IS_PYPY or not IS_X86_64,
    reason=(
        "The p537 1.0.7 release only supports CPython >=3.6,<3.13 and only has published wheels "
        "for Linux and Mac x86_64"
    ),
)


def interpreter_tag(interpreter=None):
    # type: (Optional[PythonInterpreter]) -> Tag
    return (interpreter or PythonInterpreter.get()).platform.tag


def python_tag(interpreter=None):
    # type: (Optional[PythonInterpreter]) -> str
    tag = interpreter_tag(interpreter=interpreter)
    return cast(str, tag.interpreter)


def abi_tag(interpreter=None):
    # type: (Optional[PythonInterpreter]) -> str
    tag = interpreter_tag(interpreter=interpreter)
    return cast(str, tag.abi)


@skip_unless_p537_compatible
def test_sync_strict_to_strict(tmpdir):
    # type: (Any) -> None

    # Make sure we can migrate to a new Python with a sync.

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock", "sync", "--style", "strict", "p537==1.0.7", "--indent", "2", "--lock", lock
    ).assert_success()
    p537_current = assert_p537_lock(
        lock, LockStyle.STRICT, expected_python_tag=python_tag(), expected_abi_tag=abi_tag()
    )
    assert len(p537_current.artifacts_by_tag) == 1

    other_python = (
        ensure_python_interpreter(PY39)
        if sys.version_info[:2] == (3, 10)
        else ensure_python_interpreter(PY310)
    )
    run_pex3(
        "lock",
        "sync",
        "--python",
        other_python,
        "--style",
        "strict",
        "p537==1.0.7",
        "--indent",
        "2",
        "--lock",
        lock,
    ).assert_success()
    other_interpreter = PythonInterpreter.from_binary(other_python)
    p537_other = assert_p537_lock(
        lock,
        LockStyle.STRICT,
        expected_python_tag=python_tag(other_interpreter),
        expected_abi_tag=abi_tag(other_interpreter),
    )
    assert len(p537_other.artifacts_by_tag) == 1


@skip_unless_p537_compatible
def test_sync_strict_to_sources(tmpdir):
    # type: (Any) -> None

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock", "sync", "--style", "strict", "p537==1.0.7", "--indent", "2", "--lock", lock
    ).assert_success()
    p537_strict = assert_p537_lock(
        lock, LockStyle.STRICT, expected_python_tag=python_tag(), expected_abi_tag=abi_tag()
    )
    assert len(p537_strict.artifacts_by_tag) == 1
    p537_strict_tag, p537_strict_wheel = next(iter(p537_strict.artifacts_by_tag.items()))

    run_pex3(
        "lock", "sync", "--style", "sources", "p537==1.0.7", "--indent", "2", "--lock", lock
    ).assert_success()
    p537_sources = assert_p537_lock(
        lock, LockStyle.SOURCES, expected_python_tag=python_tag(), expected_abi_tag=abi_tag()
    )
    # We should have kept the pin but picked up an additional source artifact.
    assert p537_strict.pin == p537_sources.pin
    assert len(p537_sources.artifacts_by_tag) == 2
    assert p537_strict_wheel == p537_sources.artifacts_by_tag[p537_strict_tag]
    assert (
        None in p537_sources.artifacts_by_tag
    ), "Expected the `--style sources` lock to contain an sdist"


def test_sync_universal_to_universal(tmpdir):
    # type: (Any) -> None

    # Make sure we can change ICs to migrate to a new Python version with a sync.

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "sync",
        "--style",
        "universal",
        "--interpreter-constraint",
        "CPython==3.10.*",
        "p537==1.0.7",
        "--indent",
        "2",
        "--lock",
        lock,
    ).assert_success()
    p537_py310 = assert_p537_lock(
        lock,
        LockStyle.UNIVERSAL,
        expected_requires_python="CPython==3.10.*",
        expected_python_tag="cp310",
        expected_abi_tag="cp310",
    )

    # We expect a wheel each for Linux, Mac and Windows as well as an sdist.
    assert len(p537_py310.artifacts_by_tag) == 4
    p537_py310_sdist = p537_py310.artifacts_by_tag[None]

    run_pex3(
        "lock",
        "sync",
        "--style",
        "universal",
        "--interpreter-constraint",
        "CPython==3.11.*",
        "p537==1.0.7",
        "--indent",
        "2",
        "--lock",
        lock,
    ).assert_success()
    p537_py311 = assert_p537_lock(
        lock,
        LockStyle.UNIVERSAL,
        expected_requires_python="CPython==3.11.*",
        expected_python_tag="cp311",
        expected_abi_tag="cp311",
    )
    assert p537_py310.pin == p537_py311.pin
    assert len(p537_py311.artifacts_by_tag) == 4
    assert p537_py310_sdist == p537_py311.artifacts_by_tag[None]


@pytest.mark.skipif(PY_VER < (3, 6), reason="The shiv 1.0.5 release requires Python >=3.6.")
def test_sync_extras(tmpdir):
    # type: (Any) -> None

    # Verify requirements with extras work with the lock update constraint system.

    lock = os.path.join(str(tmpdir), "lock.json")

    def collect_locked_requirements():
        # type: () -> Dict[ProjectName, LockedRequirement]
        locked_resolves = json_codec.load(lock).locked_resolves
        assert len(locked_resolves) == 1
        return {
            locked_requirement.pin.project_name: locked_requirement
            for locked_requirement in locked_resolves[0].locked_requirements
        }

    run_pex3("lock", "sync", "shiv==1.0.5", "--indent", "2", "--lock", lock).assert_success()
    original_locked_requirements = collect_locked_requirements()
    assert ProjectName("sphinx-click") not in original_locked_requirements

    run_pex3("lock", "sync", "shiv[rtd]==1.0.5", "--indent", "2", "--lock", lock).assert_success()
    updated_locked_requirements = collect_locked_requirements()

    for project_name, locked_requirement in original_locked_requirements.items():
        assert locked_requirement == updated_locked_requirements.pop(project_name)

    assert ProjectName("sphinx-click") in updated_locked_requirements



# Copyright 2019 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import errno
import functools
import multiprocessing
import os
import subprocess
import time
from abc import abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from threading import BoundedSemaphore, Event, Thread

from pex.common import pluralize
from pex.compatibility import Queue, cpu_count
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, Generic, cast

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        DefaultDict,
        Iterable,
        Iterator,
        List,
        Optional,
        Protocol,
        Text,
        Tuple,
        TypeVar,
        Union,
    )

    import attr  # vendor:skip

    _I = TypeVar("_I")
    _O = TypeVar("_O")
    _T = TypeVar("_T")
    _S = TypeVar("_S")
    _SE = TypeVar("_SE")
    _JE = TypeVar("_JE")
else:
    from pex.third_party import attr


class Job(object):
    """Represents a job spawned as a subprocess.

    Presents a similar API to `subprocess` except where noted.
    """

    class Error(Exception):
        """Indicates that a Job exited non-zero."""

        def __init__(
            self,
            pid,  # type: int
            command,  # type: Tuple[str, ...]
            exitcode,  # type: int
            stderr,  # type: Optional[Text]
            message,  # type: str
        ):
            # type: (...) -> None
            super(Job.Error, self).__init__(message)
            self.pid = pid
            self.command = command
            self.exitcode = exitcode
            self.stderr = stderr

    def __init__(
        self,
        command,  # type: Iterable[str]
        process,  # type: subprocess.Popen
        finalizer=None,  # type: Optional[Callable[[int], None]]
    ):
        # type: (...) -> None
        """
        :param command: The command used to spawn the job process.
        :param process: The spawned process handle.
        :param finalizer: An optional cleanup function to call exactly once with the process return
                          code when the underlying process terminates in the course of calling this
                          job's public methods.
        """
        self._command = tuple(command)
        self._process = process
        self._finalizer = finalizer

    def wait(self):
        # type: () -> None
        """Waits for the job to complete.

        :raises: :class:`Job.Error` if the job exited non-zero.
        """
        try:
            _, stderr = self._process.communicate()
            self._check_returncode(stderr)
        finally:
            self._finalize_job()

    def communicate(self, input=None):
        # type: (Optional[bytes]) -> Tuple[bytes, bytes]
        """Communicates with the job sending any input data to stdin and collecting stdout and
        stderr.

        :param input: Data to send to stdin of the job as per the `subprocess` API.
        :return: A tuple of the job's stdout and stderr as per the `subprocess` API.
        :raises: :class:`Job.Error` if the job exited non-zero.
        """
        try:
            stdout, stderr = self._process.communicate(input=input)
            self._check_returncode(stderr)
            return stdout, stderr
        finally:
            self._finalize_job()

    def kill(self):
        # type: () -> None
        """Terminates the job if it is still running.

        N.B.: This method is idempotent.
        """
        try:
            self._process.kill()
        except OSError as e:
            if e.errno != errno.ESRCH:
                raise e
        finally:
            self._finalize_job()

    def create_error(
        self,
        msg,  # type: str
        stderr=None,  # type: Optional[bytes]
    ):
        # type: (...) -> Job.Error
        """Creates an error with this Job's details.

        :param msg: The message for the error.
        :param stderr: Any stderr output captured from the job.
        :return: A job error.
        """
        err = None
        if stderr:
            err = stderr.decode("utf-8")
            msg += "\nSTDERR:\n{}".format(err)
        raise self.Error(
            pid=self._process.pid,
            command=self._command,
            exitcode=self._process.returncode,
            stderr=err,
            message=msg,
        )

    def _finalize_job(self):
        if self._finalizer is not None:
            self._finalizer(self._process.returncode)
            self._finalizer = None

    def _check_returncode(self, stderr=None):
        # type: (Optional[bytes]) -> None
        if self._process.returncode != 0:
            msg = "Executing {} failed with {}".format(
                " ".join(self._command), self._process.returncode
            )
            raise self.create_error(msg, stderr=stderr)

    def __str__(self):
        # type: () -> str
        return "pid: {pid} -> {command}".format(
            pid=self._process.pid, command=" ".join(self._command)
        )


class SpawnedJob(Generic["_T"]):
    """A handle to a spawned :class:`Job` and its associated result."""

    @classmethod
    def completed(cls, result):
        # type: (_T) -> SpawnedJob[_T]
        """Wrap an already completed result in a SpawnedJob.

        The returned job will no-op when `kill` is called since the job is already completed.

        :param result: The completed result.
        :return: A spawned job whose result is already complete.
        """

        class Completed(SpawnedJob):
            def await_result(self):
                # type: () -> _T
                return result

            def kill(self):
                # type: () -> None
                pass

            def __repr__(self):
                # type: () -> str
                return "SpawnedJob.completed({!r})".format(result)

        return Completed()

    @classmethod
    def wait(
        cls,
        job,  # type: Job
        result,  # type: _T
    ):
        # type: (...) -> SpawnedJob[_T]
        """Wait for the job to complete and return a fixed result upon success.

        :param job: The spawned job.
        :param result: The fixed success result.
        :return: A spawned job whose result is a side effect of the job (a written file, a populated
                 directory, etc.).
        """
        return cls.and_then(job, lambda: result)

    @classmethod
    def and_then(
        cls,
        job,  # type: Job
        result_func,  # type: Callable[[], _T]
    ):
        # type: (...) -> SpawnedJob[_T]
        """Wait for the job to complete and return a result derived from its side effects.

        :param job: The spawned job.
        :param result_func: A function that will be called to produce the result upon job success.
        :return: A spawned job whose result is derived from a side effect of the job (a written
                 file, a populated directory, etc.).
        """

        class AndThen(SpawnedJob):
            def await_result(self):
                # type: () -> _T
                job.wait()
                return result_func()

            def kill(self):
                # type: () -> None
                job.kill()

            def __repr__(self):
                # type: () -> str
                return "SpawnedJob.and_then({!r})".format(job)

        return AndThen()

    @classmethod
    def stdout(
        cls,
        job,  # type: Job
        result_func,  # type: Callable[[bytes], _T]
        input=None,  # type: Optional[bytes]
    ):
        # type: (...) -> SpawnedJob[_T]
        """Wait for the job to complete and return a result derived from its stdout.

        :param job: The spawned job.
        :param result_func: A function taking the stdout byte string collected from the spawned job
                            and returning the desired result.
        :param input: Optional input stream data to pass to the process as per the
                      `subprocess.Popen.communicate` API.
        :return: A spawned job whose result is derived from stdout contents.
        """

        class Stdout(SpawnedJob):
            def await_result(self):
                # type: () -> _T
                stdout, _ = job.communicate(input=input)
                return result_func(stdout)

            def kill(self):
                # type: () -> None
                job.kill()

            def __repr__(self):
                # type: () -> str
                return "SpawnedJob.stdout({!r})".format(job)

        return Stdout()

    @classmethod
    def file(
        cls,
        job,  # type: Job
        output_file,  # type: str
        result_func,  # type: Callable[[bytes], _T]
        input=None,  # type: Optional[bytes]
    ):
        # type: (...) -> SpawnedJob[_T]
        """Wait for the job to complete and return a result derived from a file the job creates.

        :param job: The spawned job.
        :param output_file: The path of the file the job will create.
        :param result_func: A function taking the byte contents of the file the spawned job
                            created and returning the desired result.
        :param input: Optional input stream data to pass to the process as per the
                      `subprocess.Popen.communicate` API.
        :return: A spawned job whose result is derived from the contents of a file it creates.
        """

        def _read_file(stderr=None):
            # type: (Optional[bytes]) -> bytes
            try:
                with open(output_file, "rb") as fp:
                    return fp.read()
            except (OSError, IOError) as e:
                raise job.create_error(
                    "Expected job to create file {output_file!r} but it did not exist or could not "
                    "be read: {err}".format(output_file=output_file, err=e),
                    stderr=stderr,
                )

        class File(SpawnedJob):
            def await_result(self):
                # type: () -> _T
                _, stderr = job.communicate(input=input)
                return result_func(_read_file(stderr=stderr))

            def kill(self):
                # type: () -> None
                job.kill()

            def __repr__(self):
                # type: () -> str
                return "SpawnedJob.file({job!r}, output_file={output_file!r})".format(
                    job=job, output_file=output_file
                )

        return File()

    def await_result(self):
        # type: () -> _T
        """Waits for the spawned job to complete and returns its result."""
        raise NotImplementedError()

    def kill(self):
        # type: () -> None
        """Terminates the spawned job if it's not already complete."""
        raise NotImplementedError()

    def map(self, func):
        # type: (Callable[[_T], _S]) -> SpawnedJob[_S]

        class Map(SpawnedJob):
            def await_result(me):
                # type: () -> _S
                return func(self.await_result())

            def kill(me):
                # type: () -> None
                self.kill()

            def __repr__(me):
                # type: () -> str
                return "{job}.map({func})".format(job=self, func=func)

        return Map()


# If `cpu_count` fails, we default to 2. This is relatively arbitrary, based on what seems to be
# common in CI.
_CPU_COUNT = cpu_count() or 2
_ABSOLUTE_MAX_JOBS = _CPU_COUNT * 2


DEFAULT_MAX_JOBS = _CPU_COUNT
"""The default maximum number of parallel jobs PEX should use."""


def _sanitize_max_jobs(max_jobs=None):
    # type: (Optional[int]) -> int
    if max_jobs is None or max_jobs <= 0:
        return DEFAULT_MAX_JOBS
    else:
        return min(max_jobs, _ABSOLUTE_MAX_JOBS)


class ErrorHandler(Generic["_I", "_SE", "_JE"]):
    """Handles errors encountered in the context of spawning and awaiting the result of a `Job`."""

    @classmethod
    def spawn_error_message(
        cls,
        item,  # type: _I
        exception,  # type: Exception
    ):
        # type: (...) -> str
        return "Failed to spawn a job for {item}: {exception}".format(
            item=item, exception=exception
        )

    @classmethod
    def job_error_message(
        cls,
        _item,  # type: _I
        job_error,  # type: Job.Error
    ):
        # type: (...) -> str
        return "pid {pid} -> {command} exited with {exitcode} and STDERR:\n{stderr}".format(
            pid=job_error.pid,
            command=" ".join(job_error.command),
            exitcode=job_error.exitcode,
            stderr=job_error.stderr,
        )

    @abstractmethod
    def handle_spawn_error(
        self,
        item,  # type: _I
        exception,  # type: Exception
    ):
        # type: (...) -> _SE
        """Handle an error encountered spawning a job.

        :param item: The item that was the input for the spawned job.
        :param exception: The exception encountered attempting to spawn the job for `item`.
        :type exception: :class:`Exception`
        :returns: A value to represent the failed processing of the item or else `None` to skip
                  processing of the item altogether.
        :raise: To indicate all item processing should be cancelled and the exception raised.
        """

    @abstractmethod
    def handle_job_error(
        self,
        item,  # type: _I
        job_error,  # type: Job.Error
    ):
        # type: (...) -> _JE
        """Handle a job that exits unsuccessfully.

        :param item: The item that was the input for the spawned job.
        :param job_error: An error capturing the details of the job failure.
        :type job_error: :class:`Job.Error`
        :returns: A value to represent the failed processing of the item or else `None` to skip
                  processing of the item altogether.
        :raise: To indicate all item processing should be cancelled and the exception raised.
        """


class Raise(ErrorHandler["_I", "_O", "_O"], Generic["_I", "_O"]):
    """Re-raises errors encountered spawning or awaiting the result of a `Job`."""

    def __init__(self, raise_type):
        """
        :param raise_type: The type of exception to raise when a `Job` fails.
        :type raise_type: An :class:`Exception` subclass.
        """
        self._raise_type = raise_type

    def handle_spawn_error(self, item, exception):
        raise self._raise_type(self.spawn_error_message(item, exception))

    def handle_job_error(self, item, job_error):
        raise self._raise_type(self.job_error_message(item, job_error))


class Retain(ErrorHandler["_I", "Tuple[_I, Exception]", "Tuple[_I, Job.Error]"], Generic["_I"]):
    """Retains errors encountered spawning or awaiting the result of a `Job`.

    The retained errors are returned as the result of the failed `Job` in the form of (item, error)
    tuples. In the case of a spawn failure, the error item is likely an instance of `OSError`. In
    the case of the `Job` failing (exiting non-zero), the error will be an instance of `Job.Error`.
    """

    def handle_spawn_error(self, item, exception):
        return item, exception

    def handle_job_error(self, item, job_error):
        return item, job_error


class Log(ErrorHandler["_I", "_O", "_O"], Generic["_I", "_O"]):
    """Logs errors encountered spawning or awaiting the result of a `Job`."""

    def handle_spawn_error(self, item, exception):
        TRACER.log(self.spawn_error_message(item, exception))
        return None

    def handle_job_error(self, item, job_error):
        TRACER.log(self.job_error_message(item, job_error))
        return None


def execute_parallel(
    inputs,  # type: Iterable[_I]
    spawn_func,  # type: Callable[[_I], SpawnedJob[_O]]
    error_handler=None,  # type: Optional[ErrorHandler[_I, _SE, _JE]]
    max_jobs=None,  # type: Optional[int]
):
    # type: (...) -> Iterator[Union[_O, _SE, _JE]]
    """Execute jobs for the given inputs in parallel subprocesses.

    :param int max_jobs: The maximum number of parallel jobs to spawn.
    :param inputs: An iterable of the data to parallelize over `spawn_func`.
    :param spawn_func: A function taking a single input and returning a :class:`SpawnedJob`.
    :param error_handler: An optional :class:`ErrorHandler`, defaults to :class:`Log`.
    :returns: An iterator over the spawned job results as they come in.
    :raises: A `raise_type` exception if any individual job errors and `raise_type` is not `None`.
    """
    handler = (
        error_handler or Log["_I", "_O"]()
    )  # type: Union[ErrorHandler[_I, _SE, _JE], Log[_I, _O]]
    size = _sanitize_max_jobs(max_jobs)
    TRACER.log(
        "Spawning a maximum of {} parallel jobs to process:\n  {}".format(
            size, "\n  ".join(map(str, inputs))
        ),
        V=9,
    )

    @attr.s(frozen=True)
    class Spawn(object):
        item = attr.ib()  # type: Any
        spawned_job = attr.ib()  # type: SpawnedJob

    @attr.s(frozen=True)
    class SpawnError(object):
        item = attr.ib()  # type: Any
        error = attr.ib()  # type: Exception

    stop = Event()  # Used as a signal to stop spawning further jobs once any one job fails.
    job_slots = BoundedSemaphore(value=size)

    class DoneSentinel(object):
        pass

    done_sentinel = DoneSentinel()
    spawn_queue = Queue()  # type: Queue[Union[Spawn, SpawnError, DoneSentinel]]

    def spawn_jobs():
        for item in inputs:
            if stop.is_set():
                break
            job_slots.acquire()
            try:
                result = Spawn(item, spawn_func(item))
            except Exception as e:
                result = SpawnError(item, e)
            finally:
                spawn_queue.put(result)
        spawn_queue.put(done_sentinel)

    @contextmanager
    def spawned_jobs():
        spawner = Thread(name="PEX Parallel Job Spawner", target=spawn_jobs)
        spawner.daemon = True
        spawner.start()
        try:
            yield
        finally:
            stop.set()
            # N.B.: We want to ensure, no matter what, the spawn_jobs loop above spins at least once
            # so that it can see stop is set and exit in the case it is currently blocked on a put
            # waiting for a job slot.
            try:
                job_slots.release()
            except ValueError:
                # From the BoundedSemaphore doc:
                #
                #   If the number of releases exceeds the number of acquires,
                #   raise a ValueError.
                #
                # In the normal case there will be no job_slots to release; so we expect the
                # BoundedSemaphore to raise here. We're guarding against the abnormal case where
                # there is a bug in the state machine implied by the execute_parallel code.
                pass
            spawner.join()

    with spawned_jobs():
        error = None
        while True:
            spawn_result = spawn_queue.get()

            if isinstance(spawn_result, DoneSentinel):
                if error:
                    raise error
                return

            try:
                if isinstance(spawn_result, SpawnError):
                    try:
                        se_result = handler.handle_spawn_error(
                            spawn_result.item, spawn_result.error
                        )
                        if se_result is not None:
                            yield se_result
                    except Exception as e:
                        # Fail fast and proceed to kill all outstanding spawned jobs.
                        stop.set()
                        error = e
                elif (
                    error is not None
                ):  # I.E.: `item` is not an exception, but there was a prior exception.
                    spawn_result.spawned_job.kill()
                else:
                    try:
                        yield spawn_result.spawned_job.await_result()
                    except Job.Error as e:
                        try:
                            je_result = handler.handle_job_error(spawn_result.item, e)
                            if je_result is not None:
                                yield je_result
                        except Exception as e:
                            # Fail fast and proceed to kill all outstanding spawned jobs.
                            stop.set()
                            error = e
            finally:
                job_slots.release()


if TYPE_CHECKING:

    class Comparable(Protocol):
        def __lt__(self, other):
            # type: (Any) -> bool
            pass


now = getattr(time, "perf_counter", getattr(time, "clock", time.time))  # type: Callable[[], float]


def _apply_function(
    function,  # type: Callable[[_I], _O]
    input_item,  # type: _I
):
    # type: (...) -> Tuple[int, _O, float]
    start = now()
    result = function(input_item)
    return os.getpid(), result, now() - start


if TYPE_CHECKING:

    class Pool(Protocol):
        def imap_unordered(
            self,
            func,  # type: Callable[[_I], _O]
            iterable,  # type: Iterable[_I]
            chunksize=1,  # type: int
        ):
            # type: (...) -> Iterator[_O]
            pass

        def close(self):
            # type: () -> None
            pass

        def join(self):
            # type: () -> None
            pass


@contextmanager
def _mp_pool(size):
    # type: (int) -> Iterator[Pool]
    try:
        context = multiprocessing.get_context("fork")  # type: ignore[attr-defined]
        pool = cast("Pool", context.Pool(processes=size))
    except (AttributeError, ValueError):
        pool = multiprocessing.Pool(processes=size)

    try:
        yield pool
    finally:
        pool.close()
        pool.join()


# This is derived from experiment and backed up by multiprocessing internal default chunk size of
# 4 for chunked mapping. The overhead of setting up and bookkeeping the multiprocessing pool
# processes and communication pipes seems to be worth if at least 4 items are processed per slot.
MULTIPROCESSING_DEFAULT_MIN_AVERAGE_LOAD = 4


def iter_map_parallel(
    inputs,  # type: Iterable[_I]
    function,  # type: Callable[[_I], _O]
    max_jobs=None,  # type: Optional[int]
    min_average_load=MULTIPROCESSING_DEFAULT_MIN_AVERAGE_LOAD,  # type: int
    costing_function=None,  # type: Optional[Callable[[_I], Comparable]]
    result_render_function=None,  # type: Optional[Callable[[_O], Any]]
    noun="item",  # type: str
    verb="process",  # type: str
    verb_past="processed",  # type: str
):
    # type: (...) -> Iterator[_O]
    """Enhanced `multiprocessing.Pool.imap_unordered` that optimizes pool size and input ordering.

    :param inputs: The items to process with `function`.
    :param function: A function that takes a single argument from `inputs` and returns a result.
    :param max_jobs: The maximum number of Python processes to spawn to service the `inputs`.
    :param min_average_load: The minimum avg. number of inputs each Python process should service.
    :param costing_function: A function that can estimate the cost of processing each input.
    :param result_render_function: A function that can take a result from `function` and render an
                                   identifier for it.
    :param noun: A noun indicating what the input type is; "item" by default.
    :param verb: A verb indicating what the function does; "process" by default.
    :param verb_past: The past tense of `verb`; "processed" by default.
    :return: An iterator over the mapped results.
    """
    input_items = list(inputs)
    if not input_items:
        return

    if costing_function is not None:
        # We ensure no job slot is so unlucky as to get all the biggest jobs and thus become an
        # un-necessarily long pole by sorting based on cost. Some examples to illustrate the effect
        # using 6 input items and 2 job slots:
        #
        # 1.) Random worst case ordering:
        #         [9, 1, 1, 1, 1, 10] -> slot1[9] slot2[1, 1, 1, 1, 10]: 14 long pole
        #     Sorted becomes:
        #         [10, 9, 1, 1, 1, 1] -> slot1[10, 1, 1] slot2[9, 1, 1]: 12 long pole
        # 2.) Random worst case ordering:
        #         [6, 4, 3, 10, 1, 1] -> slot1[6, 10] slot2[4, 3, 1, 1]: 16 long pole
        #     Sorted becomes:
        #         [10, 6, 4, 3, 1, 1] -> slot1[10, 3] slot2[6, 4, 1, 1]: 13 long pole
        #
        input_items.sort(key=costing_function, reverse=True)

    # We want each of the job slots above to process MULTIPROCESSING_MIN_AVERAGE_LOAD on average in
    # order to overcome multiprocessing overheads. Of course, if there are fewer available cores
    # than that or the user has pinned max jobs lower, we clamp to that. Finally, we always want at
    # least two slots to ensure we process input items in parallel.
    pool_size = max(2, min(len(input_items) // min_average_load, _sanitize_max_jobs(max_jobs)))

    apply_function = functools.partial(_apply_function, function)

    slots = defaultdict(list)  # type: DefaultDict[int, List[float]]
    with TRACER.timed(
        "Using {pool_size} parallel jobs to {verb} {count} items".format(
            pool_size=pool_size, verb=verb, count=len(input_items)
        )
    ):
        with _mp_pool(size=pool_size) as pool:
            for pid, result, elapsed_secs in pool.imap_unordered(apply_function, input_items):
                TRACER.log(
                    "[{pid}] {verbed} {result} in {elapsed_secs:.2f}s".format(
                        pid=pid,
                        verbed=verb_past,
                        result=result_render_function(result) if result_render_function else result,
                        elapsed_secs=elapsed_secs,
                    ),
                    V=2,
                )
                yield result
                slots[pid].append(elapsed_secs)

    TRACER.log(
        "Elapsed time per {verb} job:\n  {times}".format(
            verb=verb,
            times="\n  ".join(
                "{index}) [{pid}] {total_secs:.2f}s {count} {inputs}".format(
                    index=index,
                    pid=pid,
                    count=len(elapsed),
                    inputs=pluralize(elapsed, noun),
                    total_secs=total_secs,
                )
                for index, (total_secs, pid, elapsed) in enumerate(
                    sorted(
                        ((sum(elapsed), pid, elapsed) for pid, elapsed in slots.items()),
                        reverse=True,
                    ),
                    start=1,
                )
            ),
        )
    )


def map_parallel(
    inputs,  # type: Iterable[_I]
    function,  # type: Callable[[_I], _O]
    max_jobs=None,  # type: Optional[int]
    min_average_load=MULTIPROCESSING_DEFAULT_MIN_AVERAGE_LOAD,  # type: int
    costing_function=None,  # type: Optional[Callable[[_I], Comparable]]
    result_render_function=None,  # type: Optional[Callable[[_O], Any]]
    noun="item",  # type: str
    verb="process",  # type: str
    verb_past="processed",  # type: str
):
    # type: (...) -> List[_O]
    """Enhanced version of `multiprocessing.Pool.map` that optimizes pool size and input ordering.

    Unlike `multiprocessing.Pool.map`, the output order is not guaranteed.

    Forwards all arguments to `imap_parallel`.

    :return: A list of the mapped results.
    """
    return list(
        iter_map_parallel(
            inputs,
            function,
            max_jobs=max_jobs,
            min_average_load=min_average_load,
            costing_function=costing_function,
            result_render_function=result_render_function,
            noun=noun,
            verb=verb,
            verb_past=verb_past,
        )
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import shutil
import sys

from pex.typing import TYPE_CHECKING
from testing import PY39, data, ensure_python_interpreter, run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any


def test_download_incompatible_python(tmpdir):
    # type: (Any) -> None

    python = ensure_python_interpreter(PY39) if sys.version_info >= (3, 10) else sys.executable

    pex_root = os.path.join(str(tmpdir), "pex_root")
    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "--pex-root",
        pex_root,
        "--style",
        "universal",
        "--interpreter-constraint",
        "CPython==3.11.*",
        "pbipy==2.8.2",
        "--indent",
        "2",
        "-o",
        lock,
        python=python,
    ).assert_success()

    complete_platform = data.path("platforms", "complete_platform_linux_x86-64_py11.json")
    pex = os.path.join(str(tmpdir), "pex")
    shutil.rmtree(pex_root)
    run_pex_command(
        args=[
            "--pex-root",
            pex_root,
            "--complete-platform",
            complete_platform,
            "pbipy==2.8.2",
            "--lock",
            lock,
            "-o",
            pex,
        ],
        python=python,
    ).assert_success()
